import numpy as np
from scipy import sparse
import warnings
import inspect

from ..problems.composite import smooth as smooth_composite
from ..affine import affine_transform, linear_transform, astransform
from ..identity_quadratic import identity_quadratic

#TODO: create proximal methods for known smooth things
class smooth_atom(smooth_composite):

    """
    A class for representing a smooth function and its gradient

    Parameters
    ----------

    shape : tuple
       Shape of argument to `smooth_objective`

    coef : float (optional)
       Scalar multiple to be applied (must be nonnegative)

    offset : ndarray (optional)
       Vector to be subtracted before evaluating `smooth_objective`. 

    quadratic : `identity_quadratic` (optional)
       Instance of `identity_quadratic` to be added to overall
       objective.

    initial : ndarray (optional)
       Initial value for coefficients.

    """

    objective_template = r'''f(%(var)s)'''
    objective_vars = {'var':r'\beta',
                      'shape':'p',
                      'coef':'C',
                      'offset':r'\alpha+'}

    def __init__(self, shape, coef=1, offset=None,
                 quadratic=None, initial=None):
        smooth_composite.__init__(self, shape,
                                  offset=offset,
                                  quadratic=quadratic,
                                  initial=initial)
        self.coef = coef
        if coef < 0:
            raise ValueError('coefs must be nonnegative to ensure convexity (assuming all atoms are indeed convex)')
        self.coefs = np.zeros(self.shape)

    def smooth_objective(self, arg, mode='both', check_feasibility=False):
        """

        Parameters
        ----------

        arg : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `beta` is not
            in the domain.

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `beta`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """
        raise NotImplementedError("Abstract method.")
    
    @classmethod
    def affine(cls, linear_operator, offset, coef=1, diag=False,
               quadratic=None, **kws):
        """
        Keywords given in kws are passed to cls constructor along with other arguments
        """
        if not isinstance(linear_operator, affine_transform):
            l = linear_transform(linear_operator, diag=diag)
        else:
            l = linear_operator
        if not acceptable_init_args(cls, kws):
            raise ValueError("Invalid arguments being passed to initialize " + cls.__name__)
        
        # the minus signs below for offset is there until affine transforms SUBTRACT 
        # their offset until add. 
        # for atoms, the offset is really the "center"

        atom = cls(l.output_shape, coef=coef, offset=-offset, quadratic=quadratic, **kws)
        
        return affine_smooth(atom, l)

    @classmethod
    def linear(cls, linear_operator, coef=1, diag=False,
               offset=None, 
               quadratic=None, **kws):
        """
        Keywords given in kws are passed to cls constructor along with other arguments
        """
        if not acceptable_init_args(cls, kws):
            raise ValueError("Invalid arguments being passed to initialize " + cls.__name__)

        atransform = affine_transform(linear_operator, None, diag=diag)
        atom = cls(atransform.output_shape, coef=coef, quadratic=quadratic, offset=offset, **kws)
        
        return affine_smooth(atom, atransform)

    @classmethod
    def shift(cls, offset, coef=1, quadratic=None, **kws):
        """
        Keywords given in kws are passed to cls constructor along with other arguments
        """
        if not acceptable_init_args(cls, kws):
            raise ValueError("Invalid arguments being passed to initialize " + cls.__name__)
        
        atom = cls(offset.shape, coef=coef, quadratic=quadratic, 
                   offset=offset, **kws)
        return atom

    def scale(self, obj, copy=False):
        if self.coef != 1:
            return obj * self.coef
        if copy:
            return obj.copy()
        return obj

    def get_conjugate(self):
        raise NotImplementedError('each smooth loss should implement its own get_conjugate')

    @property
    def conjugate(self):
        return self.get_conjugate()
 

def acceptable_init_args(cls, proposed_keywords):
    """
    Check that the keywords in the dictionary proposed_keywords are arguments to __init__ of class cls

    Returns True/False
    """
    args = inspect.getargspec(cls.__init__).args
    forbidden = ['self', 'shape', 'coef', 'quadratic', 'initial', 'offset']
    for kw in proposed_keywords.keys():
        if not kw in args:
            return False
        if kw in forbidden:
            return False
    return True

class affine_smooth(smooth_atom):

    """

    Composition of a smooth objective with an affine transform.

    """

    force_reshape = True

    objective_vars = {'linear':'X'}

    def __init__(self, smooth_atom, atransform, store_grad=True, diag=False):
        """

        Parameters
        ----------

        smooth_atom : `regreg.smooth.smooth_atom`
             A smooth atom.

        atransform : `regreg.affine.affine_transform`
             An affine transformation, or cast to one
             using `regreg.affine.linear_transform`

        store_grad : bool
             If True, when computing the gradient,
             store a reference to the gradient of `smooth_atom`
             in the attribute `grad`.

        diag : bool
             Indicates if `atransform` acts diagonally,
             i.e. a rescaling.
             Passed to `regreg.affine.linear_transform`.

        """
        self.store_grad = store_grad
        self.atom = smooth_atom
        if not isinstance(atransform, affine_transform):
            atransform = linear_transform(atransform, diag=diag)
        self.affine_transform = atransform
        self.shape = atransform.input_shape
        self.coefs = np.zeros(self.shape)

    def _get_coef(self):
        return self.atom.coef

    def _set_coef(self, coef):
        self.atom.coef = coef
    coef = property(_get_coef, _set_coef)

    def smooth_objective(self, arg, mode='both', check_feasibility=False):
        """
        Compute the smooth objective at the point `self.transform.affine_map(arg)`.

        Parameters
        ----------

        arg : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `beta` is not
            in the domain.

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `self.transform(arg)`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """

        eta = self.affine_transform.affine_map(arg)
        if mode == 'both':
            v, g = self.atom.smooth_objective(eta, mode='both')
            if self.store_grad:
                self.grad = g
            g = self.affine_transform.adjoint_map(g)
            if self.force_reshape:
                g = g.reshape(self.shape)
            return v, g
        elif mode == 'grad':
            g = self.atom.smooth_objective(eta, mode='grad')
            if self.store_grad:
                self.grad = g
            g = self.affine_transform.adjoint_map(g)
            if self.force_reshape:
                g = g.reshape(self.shape)
            return g 
        elif mode == 'func':
            v = self.atom.smooth_objective(eta, mode='func')
            return v 

    @property
    def dual(self):
        try: 
            conj = self.atom.conjugate
            return self.affine_transform, conj
        except:
            return None

    def __repr__(self):
        return ("affine_smooth(%s, %s, store_grad=%s)" % 
                (str(self.atom),
                str(self.affine_transform),
                self.store_grad))

    def latexify(self, var=None, idx=''):
        template_dict = self.atom.objective_vars.copy()
        template_dict['linear'] = self.objective_vars['linear']
        if var is not None:
            template_dict['var'] = var
        template_dict['idx'] = idx

        obj_latex = self.atom.latexify(var='%(linear)s_{%(idx)s}%(var)s' % template_dict, idx=idx)
        if not self.quadratic.iszero:
            return ' + '.join([obj_latex, self.quadratic.latexify(var=template_dict['var'], idx=idx)])
        else:
            return obj_latex

class zero(smooth_atom):

    """
    The zero function.
    """

    def smooth_objective(self, x, mode='both', check_feasibility=False):
        if mode == 'both':
            return 0., np.zeros(x.shape)
        elif mode == 'func':
            return 0.
        elif mode == 'grad':
            return np.zeros(x.shape)
        raise ValueError("Mode not specified correctly")

class sum(smooth_atom):
    """
    A simple way to combine smooth objectives
    """
    def __init__(self, atoms, weights=None):
        """
        Parameters
        ----------

        atoms : sequence
            A sequence of `regreg.smooth.smooth_atom` that will be summed
            to make a new atom.

        weights : ndarray (optional)
            If provided, these weights will appear as coefficients
            in front of each atom.

        """
        self.offset = None
        self.atoms = atoms
        if weights is None:
            weights = np.ones(len(self.atoms))
        self.weights = np.asarray(weights).reshape(-1)
        if self.weights.shape[0] != len(atoms):
            raise ValueError('weights and atoms have different lengths')
        if np.any(self.weights < 0):
            raise ValueError('weights should be non-negative to maintain convexity')
        self.coefs = self.atoms[0].coefs
        self.shape = self.coefs.shape

    def smooth_objective(self, x, mode='both', check_feasibility=False):

        """
        Compute the smooth objective at the point `self.transform.affine_map(arg)`,
        which is the sum of each `atom`'s objective with its respective weight.

        Parameters
        ----------

        arg : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `beta` is not
            in the domain.

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `self.transform(arg)`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """

        x = self.apply_offset(x)
        f, g = 0, 0
        for w, atom in zip(self.weights, self.atoms):
            if mode == 'func':
                f += w * atom.smooth_objective(x, 'func')
            elif mode == 'grad':
                g += w * atom.smooth_objective(x, 'grad')
            elif mode == 'both':
                fa, ga = atom.smooth_objective(x, 'both')
                f += w * fa; g += w * ga

        if mode == 'func':
            return f
        elif mode == 'grad':
            return g
        elif mode == 'both':
            return f, g
        else:
            raise ValueError("mode incorrectly specified")




"""
Solve a problem with SLOPE and compare to R
"""

from __future__ import print_function

import numpy as np
import nose.tools as nt

try:
    import rpy2.robjects as rpy
    rpy2_available = True
    import rpy2.robjects.numpy2ri
    from rpy2.robjects.packages import importr
    from rpy2 import robjects
except ImportError:
    rpy2_available = False

Rslope = True
try:
    SLOPE = importr('SLOPE')
except:
    Rslope = False
    

import regreg.api as rr
from regreg.atoms.slope import slope
from regreg.tests.decorators import set_seed_for_test

def fit_slope_R(X, Y):
    rpy2.robjects.numpy2ri.activate()
    robjects.r('''
    slope = function(X, Y, W=NA, choice_weights, tol_infeas = 1e-6,
       tol_rel_gap = 1e-6){

       result = SLOPE(X, Y, q = 0.1, lambda='bh', scale='l2',
                      intercept=FALSE,
                      tol_infeas = tol_infeas, tol_rel_gap = tol_rel_gap)
      
       print(result$alpha)
       return(list(beta = result$coefficients, E = result$selected, lambda_seq = result$lambda, sigma = result$sigma,
                   alpha=result$alpha))
    }''')

    r_slope = robjects.globalenv['slope']

    n, p = X.shape
    r_X = robjects.r.matrix(X, nrow=n, ncol=p)
    r_Y = robjects.r.matrix(Y, nrow=n, ncol=1)

    result = r_slope(r_X, r_Y)

    rpy2.robjects.numpy2ri.deactivate()

    return result[0], result[1], result[2], result[3], result[4]

@set_seed_for_test(10)
@np.testing.dec.skipif(True, msg="SLOPE parameterization in R has changed")
def test_using_SLOPE_weights():

    n, p = 500, 50

    X = np.random.standard_normal((n, p))
    #Y = np.random.standard_normal(n)
    X -= X.mean(0)[None, :]
    X /= (X.std(0)[None, :] * np.sqrt(n))
    beta = np.zeros(p)
    beta[:5] = 5.

    Y = X.dot(beta) + np.random.standard_normal(n)

    output_R = fit_slope_R(X, Y)
    r_beta = np.squeeze(output_R[0])[:,3]
    r_lambda_seq = np.array(output_R[2]).reshape(-1)
    alpha = output_R[-1]

    W = np.asarray(r_lambda_seq * alpha[3]).reshape(-1)
    pen = slope(W, lagrange=1.)

    loss = rr.squared_error(X, Y)
    problem = rr.simple_problem(loss, pen)
    soln = problem.solve(tol=1.e-14, min_its=500)

    # we get a better objective value
    nt.assert_true(problem.objective(soln) < problem.objective(np.asarray(r_beta)))
    nt.assert_true(np.linalg.norm(soln - r_beta) < 1.e-6 * np.linalg.norm(soln))


@set_seed_for_test(10)
@np.testing.dec.skipif(True, msg="SLOPE parameterization in R has changed")
def test_using_SLOPE_prox():

    n, p = 50, 50

    X = np.identity(n)
    beta = np.zeros(p)
    beta[:5] = 5.

    Y = X.dot(beta) + np.random.standard_normal(n)

    output_R = fit_slope_R(X, Y)
    r_beta = np.squeeze(output_R[0])[:,3]
    r_lambda_seq = np.array(output_R[2]).reshape(-1)
    alpha = output_R[-1]

    W = np.asarray(r_lambda_seq * alpha[3]).reshape(-1)
    pen = slope(W, lagrange=1.)

    soln = pen.lagrange_prox(Y)

    # test that the prox maps agree
    nt.assert_true(np.linalg.norm(soln - r_beta) < 1.e-10 * np.linalg.norm(soln))





import numpy as np

from .. import (lasso, 
                group_lasso,
                sparse_group_lasso,
                sparse_group_block,
                cross_validate)

from ...tests.decorators import set_seed_for_test

@set_seed_for_test()
def test_lasso(n=1000, p=100, nstep=20):
    '''
    CV for a LASSO path

    '''
    X = np.random.standard_normal((n,p))
    Y = np.random.standard_normal(n)
    betaX = np.zeros(p)
    betaX[:3] = [3,4,5]
    Y += np.dot(X, betaX) + np.random.standard_normal(n)

    lasso1 = lasso.gaussian(X, Y, np.ones(X.shape[1]))
    lagrange_sequence = lasso.default_lagrange_sequence(lasso1.penalty,
                                                        lasso1.grad_solution, # initialized at "null" model
                                                        nstep=nstep)

    print(cross_validate.cross_validate(lasso1,
                                        lagrange_sequence,
                                        inner_tol=1.e-5,
                                        cv=3))

@set_seed_for_test()
def test_group_lasso(n=1000, p=100, nstep=20):
    '''
    CV for a LASSO path

    '''
    X = np.random.standard_normal((n,p))
    Y = np.random.standard_normal(n)
    betaX = np.zeros(p)
    betaX[:3] = [3,4,5]
    np.random.shuffle(betaX)
    Y += np.dot(X, betaX) + np.random.standard_normal(n)

    groups = ['a', 'a', 'a', 'b', 'b', 'c']
    i = 0
    while True:
        groups.extend([str(i)]*5)
        if len(groups) > p:
            groups = np.asarray(groups)
            groups = groups[:p]
            break
        i += 1
    np.random.shuffle(groups)
    group_lasso1 = group_lasso.gaussian(X, 
                                        Y, 
                                        groups)
    lagrange_sequence = group_lasso.default_lagrange_sequence(group_lasso1.penalty,
                                                              group_lasso1.grad_solution,
                                                              nstep=nstep) # initialized at "null" model

    cross_validate.cross_validate(group_lasso1,
                                  lagrange_sequence,
                                  inner_tol=1.e-5,
                                  cv=3)


@set_seed_for_test()
def test_sparse_group_lasso(n=1000, p=100, nstep=20):
    '''
    CV for a LASSO path

    '''
    X = np.random.standard_normal((n,p))
    Y = np.random.standard_normal(n)
    betaX = np.zeros(p)
    betaX[:3] = [3,4,5]
    np.random.shuffle(betaX)
    Y += np.dot(X, betaX) + np.random.standard_normal(n)

    groups = ['a', 'a', 'a', 'b', 'b', 'c']
    i = 0
    while True:
        groups.extend([str(i)]*5)
        if len(groups) > p:
            groups = np.asarray(groups)
            groups = groups[:p]
            break
        i += 1
    np.random.shuffle(groups)
    l1weights = np.ones(p)
    sparse_group_lasso1 = sparse_group_lasso.gaussian(X, 
                                                      Y, 
                                                      groups,
                                                      l1weights)
    lagrange_sequence = sparse_group_lasso.default_lagrange_sequence(sparse_group_lasso1.penalty,
                                                                     sparse_group_lasso1.grad_solution, # initialized at "null" model
                                                                     nstep=nstep) 

    cross_validate.cross_validate(sparse_group_lasso1,
                                  lagrange_sequence,
                                  inner_tol=1.e-10,
                                  cv=3)



@set_seed_for_test()
def test_sparse_group_lasso(n=1000, p=100, nstep=20):
    '''
    CV for a LASSO path

    '''
    X = np.random.standard_normal((n,p))
    Y = np.random.standard_normal(n)
    betaX = np.zeros(p)
    betaX[:3] = [3,4,5]
    np.random.shuffle(betaX)
    Y += np.dot(X, betaX) + np.random.standard_normal(n)

    groups = ['a', 'a', 'a', 'b', 'b', 'c']
    i = 0
    while True:
        groups.extend([str(i)]*5)
        if len(groups) > p:
            groups = np.asarray(groups)
            groups = groups[:p]
            break
        i += 1
    np.random.shuffle(groups)
    l1weights = np.ones(p)
    sparse_group_lasso1 = sparse_group_lasso.gaussian(X, 
                                                      Y, 
                                                      groups,
                                                      l1weights)
    lagrange_sequence = sparse_group_lasso.default_lagrange_sequence(sparse_group_lasso1.penalty,
                                                                     sparse_group_lasso1.grad_solution, # initialized at "null" model
                                                                     nstep=nstep) 

    cross_validate.cross_validate(sparse_group_lasso1,
                                  lagrange_sequence,
                                  inner_tol=1.e-10,
                                  cv=3)

@set_seed_for_test()
def test_sparse_group_multiresponse(n=1000, p=100, q=3, nstep=20):
    '''
    CV for a LASSO path

    '''
    X = np.random.standard_normal((n,p))
    betaX = np.zeros((p,q))
    betaX[:3] = [3,4,5]
    np.random.shuffle(betaX)
    Y = np.dot(X, betaX) + np.random.standard_normal((n,q))

    sparse_group_block1 = sparse_group_block.multiresponse_gaussian(X, 
                                                                    Y, 
                                                                    1,
                                                                    np.sqrt(q))
    lagrange_sequence = sparse_group_block.default_lagrange_sequence(sparse_group_block1.penalty,
                                                                     sparse_group_block1.grad_solution,
                                                                     nstep=nstep) # initialized at "null" model

    cross_validate.cross_validate(sparse_group_block1,
                                  lagrange_sequence,
                                  inner_tol=1.e-10,
                                  cv=3)

@set_seed_for_test()
def test_sparse_group_block(n=1000, p=100, q=3, nstep=20):
    '''
    CV for a LASSO path

    '''
    n1, n2, n3, p, q = 1190, 1200, 1210, 100, 3
    X1 = np.random.standard_normal((n1,p))
    X2 = np.random.standard_normal((n2,p))
    X3 = np.random.standard_normal((n3,p))
    betaX = np.zeros((p,q))
    betaX[:3] = [3,4,5]
    betaX *= 0.2
    np.random.shuffle(betaX)
    Y = [np.dot(X, beta) + np.random.standard_normal(n) for
         X, beta, n in zip([X1, X2, X3],
                           betaX.T,
                           [n1, n2, n3])]
                                     

    sparse_group_block1 = sparse_group_block.stacked_gaussian([X1, X2, X3], 
                                                              Y, 
                                                              1,
                                                              np.sqrt(q))
    lagrange_sequence = sparse_group_block.default_lagrange_sequence(sparse_group_block1.penalty,
                                                                     sparse_group_block1.grad_solution,
                                                                     nstep=nstep) # initialized at "null" model

    cross_validate.cross_validate(sparse_group_block1,
                                  lagrange_sequence,
                                  inner_tol=1.e-10,
                                  cv=3)






"""
This module has a class for specifying a problem from just
a smooth function and a single penalty.
"""
from __future__ import print_function, division, absolute_import

import numpy as np, warnings

from ..problems.composite import composite
from ..affine import identity, scalar_multiply, astransform, adjoint
from ..atoms import atom
from ..atoms.cones import zero as zero_cone
from ..smooth import zero as zero_smooth, sum as smooth_sum, affine_smooth
from ..identity_quadratic import identity_quadratic
from ..algorithms import FISTA

class simple_problem(composite):
    
    def __init__(self, smooth_atom, proximal_atom):
        self.smooth_atom = smooth_atom
        self.proximal_atom = proximal_atom
        self.coefs = self.smooth_atom.coefs = self.proximal_atom.coefs

    def smooth_objective(self, x, mode='both', check_feasibility=False):
        """
        This class explicitly assumes that
        the proximal_atom has 0 for smooth_objective.
        """
        vs = self.smooth_atom.smooth_objective(x, mode, check_feasibility)
        return vs

    def nonsmooth_objective(self, x, check_feasibility=False):
        vn = self.proximal_atom.nonsmooth_objective(x, check_feasibility=check_feasibility)
        vs = self.smooth_atom.nonsmooth_objective(x, check_feasibility=check_feasibility)
        return vn + vs + self.quadratic.objective(x, 'func')

    def proximal(self, proxq):
        proxq = proxq + self.smooth_atom.quadratic + self.quadratic
        return self.proximal_atom.solve(proxq)

    @staticmethod
    def smooth(smooth_atom):
        """
        A problem with no nonsmooth part except possibly
        the quadratic of smooth_atom.

        The proximal function is (almost) a nullop.
        """
        proximal_atom = zero_cone(smooth_atom.shape)
        return simple_problem(smooth_atom, proximal_atom)

    @staticmethod
    def nonsmooth(proximal_atom):
        """
        A problem with no nonsmooth part except possibly
        the quadratic of smooth_atom.

        The proximal function is (almost) a nullop.
        """
        smooth_atom = zero_smooth(proximal_atom.shape)
        return simple_problem(smooth_atom, proximal_atom)

    def latexify(self, var=None):
        template_dict = self.objective_vars.copy()
        template_dict['prox'] =  self.proximal_atom.latexify(var=var,idx='2')
        template_dict['smooth'] = self.smooth_atom.latexify(var=var,idx='1')
        result = r'''
        \begin{aligned}
        \text{minimize}_{%(var)s} & f(%(var)s) + g(%(var)s) \\
        f(%(var)s) &= %(smooth)s \\
        g(%(var)s) &= %(prox)s \\
        \end{aligned}
        ''' % template_dict
        result = '\n'.join([s.strip() for s in result.split('\n')])
        return result

    def solve(self, quadratic=None, return_optimum=False, **fit_args):
        if quadratic is not None:
            oldq, self.quadratic = self.quadratic, self.quadratic + quadratic
        else:
            oldq = self.quadratic

        solver = FISTA(self)
        solver.composite.coefs[:] = self.coefs
        self.solver_results = solver.fit(**fit_args)
        self.final_step = solver.step

        if return_optimum:
            value = (self.objective(self.coefs), self.coefs)
        else:
            value = self.coefs
        self.quadratic = oldq
        return value

    
def gengrad(simple_problem, lipschitz, tol=1.0e-8, max_its=1000, debug=False,
            coef_stop=False):
    """
    A simple generalized gradient solver
    """
    itercount = 0
    coef = simple_problem.coefs
    value = np.inf
    while True:
        vnew, g = simple_problem.smooth_objective(coef, 'both')
        vnew += simple_problem.nonsmooth_objective(coef)
        newcoef = simple_problem.solve(identity_quadratic(lipschitz, coef, g, 0))
        if coef_stop:
            coef_stop_check = (np.linalg.norm(coef-newcoef) <= tol * 
                               np.max([np.linalg.norm(coef),
                                       np.linalg.norm(newcoef), 
                                       1]))
            if coef_stop_check:
                break
        else:
            obj_stop_check = np.fabs(value - vnew) <= tol * np.max([vnew, 1])
            if obj_stop_check:
                break
        if itercount == max_its:
            break
        if debug:
            print(itercount, vnew, value, (vnew - value) / vnew)
        value = vnew
        itercount += 1
        coef = newcoef
    return coef

def nesta(smooth_atom, proximal_atom, conjugate_atom, epsilon=None,
          tol=1.e-06, max_iters=100,
          min_iters=5, 
          coef_tol=1.e-6,
          initial_primal=None,
          initial_dual=None,
          coef_stop=False,
          quadratic=None):
    '''
    Parameters
    ----------

    smooth_atom: smooth_composite
        A smooth function, i.e. having a smooth_objective method.

    proximal_atom: 
        An atom with a proximal method.

    conjugate_atom:
        An atom that will be smoothed, by adding a quadratic to its
        conjugate.

    epsilon: np.array
        A decreasing array of positive constants for Moreau-Yosida smoothing.
        
    tol : np.float
        Tolerance to which each problem is solved is max(tol, epsilon)
    
    max_iter s: int
        Maximum number of iterations. If epsilon is not supplied,
        it is taken to be [1]*max_iters

    initial_primal, initial_dual : `np.ndarray(np.float)`
        Initial conditions for both primal and dual variables.

    coef_tol : float
        Tolerance for assessing convergence of coefficients.

    quadratic : `regreg.identity_quadratic`
        A quadratic term that is added to the total objective.

    Returns
    -------

    primal: np.array
        Primal coefficients.

    dual: np.array
        Dual coefficients.

    '''
    if epsilon is None:
        epsilon = [1]*max_iters # 2.**(-np.arange(20))

    transform, conjugate = conjugate_atom.dual
    if initial_dual is None:
        dual_coef = np.zeros(transform.output_shape) 
    else:
        dual_coef = initial_dual.copy()

    if quadratic is None:
        quadratic = identity_quadratic(0,0,0,0)

    primal_old = np.zeros(transform.output_shape)
    value = np.inf
    for idx, eps in enumerate(epsilon):
        smoothed = conjugate_atom.smoothed(identity_quadratic(eps, dual_coef, 0, 0))
        if smooth_atom is not None:
            final_smooth = smooth_sum([smooth_atom, smoothed])
        else:
            final_smooth = smoothed
        if proximal_atom is not None:
            problem = simple_problem(final_smooth, proximal_atom)
        else:
            problem = simple_problem.smooth(final_smooth)

        if idx == 0 and initial_primal is not None:
            problem.coefs[:] = initial_primal

        primal_coef = problem.solve(quadratic,tol=max(eps,tol))
        updated = problem.objective(primal_coef) + quadratic.objective(primal_coef, 'func')

        # should we have a better convergence criterion?
        if coef_stop:
            if (np.linalg.norm(dual_coef - smoothed.grad) < coef_tol * (np.linalg.norm(smoothed.grad) + 1)
                and idx >= min_iters):
                break
        else:
            if (np.fabs(updated - value) < coef_tol * (np.fabs(updated) + 1) and
                idx >= min_iters):
                break
        value = updated

        # when there's an affine transform involved
        dual_coef[:] = smoothed.grad + idx / (idx + 3.) * (smoothed.grad - dual_coef)
        if idx == max_iters-1:
            warnings.warn('problem not solved after %d iterations' % max_iters)

    return primal_coef, dual_coef

def tfocs(primal_atom, transform, dual_proximal_atom, epsilon=None,
          tol=1.e-06,
          max_iters=100,
          coef_tol=1.e-6,
          quadratic=None):
    '''

    This function is based on the setup of problems
    described in `TFOCS <http://tfocs.stanford.edu/>`_.
    Generally speaking, these are the same type of problems
    that nesta can handle, though without the additional smooth part.

    This solver is suited to solving problems of the form

    minimize_v f(v) + h(Dv+a)

    when both f and h (and hence f^* and h^*) 
    have simple proximal operators.

    Here is an example for minimum :math:`\ell_1` norm reconstruction.

    >>> import numpy as np, regreg.api as rr
    >>> np.random.seed(0)
    >>> n, p = 200, 5000

    The problem assumes Y=np.dot(X,beta) for some sparse beta.

    >>> X = np.random.standard_normal((n, p))
    >>> beta = np.zeros(p)
    >>> beta[:10] = np.arange(10)+1
    >>> Y = np.dot(X, beta)
    >>> 

    The problem is formally,

    minimize_v np.fabs(v).sum() subject to Y=np.dot(X,v)

    The :math:`\ell_1` norm is described as:

    >>> l1 = rr.l1norm(p, lagrange=1)

    The constraint is specified as

    >>> constraint = rr.zero_constraint.affine(X,-Y)

    >>> transform, zero = constraint.dual
    >>> epsilon = [0.01]*50 + [0.001]*20
    >>> primal_tfocs, dual_tfocs = rr.tfocs(l1, transform, zero, epsilon=epsilon)
    >>> np.linalg.norm(primal_tfocs - beta) < 1.e-3 * np.linalg.norm(beta)
    True

    >>> np.linalg.norm(primal_tfocs[10:]) < 1.e-3 * np.linalg.norm(primal_tfocs)
    True

    Parameters
    ----------

    primal_atom: atom
        An atom that will be smoothed,
        then composed with the transform.

    transform : affine_transform
        An affine transform for the composition.

    dual_proximal_atom: atom
        An atom with a proximal method.

    epsilon: np.array
        A decreasing array of positive constants for Moreau-Yosida smoothing.
        
    tol: np.float
        Tolerance to which each problem is solved is max(tol, epsilon)
    
    max_iters: int
        Maximum number of iterations. If epsilon is not supplied,
        it is taken to be [1]*max_iters

    initial_primal, initial_dual : `np.ndarray(np.float)`
        Initial conditions for both primal and dual variables.

    coef_tol : float
        Tolerance for assessing convergence of coefficients.

    quadratic : `regreg.identity_quadratic`
        A quadratic term that is added to the total objective.

    Returns
    -------

    primal: np.array
        Primal coefficients.

    dual: np.array
        Dual coefficients.

    
    '''
    transform = astransform(transform)

    #conjugate_atom needs a conjugate so it can be smoothed

    conjugate_atom = primal_atom.conjugate

    if epsilon is None:
        epsilon = [1]*max_iters # 2.**(-np.arange(20))

    offset = transform.affine_offset
    if offset is not None:
        dual_sq = identity_quadratic(0,0,-offset, 0)
    else:
        dual_sq = identity_quadratic(0,0,0,0)
        
    if quadratic is None:
        quadratic = identity_quadratic(0,0,0,0)

    primal_coef = np.zeros(conjugate_atom.shape)
    for idx, eps in enumerate(epsilon):
        smoothed = conjugate_atom.smoothed(identity_quadratic(eps, primal_coef, 0, 0) + quadratic)
        final_smooth = affine_smooth(smoothed, scalar_multiply(adjoint(transform), -1))
        problem = simple_problem(final_smooth, dual_proximal_atom)
        dual_coef = problem.solve(dual_sq, tol=max(eps,tol))

        # should we have a better convergence criterion?
        if (np.linalg.norm(primal_coef - final_smooth.grad) < 
            coef_tol * max(np.linalg.norm(final_smooth.grad),1)):
            break

        # when there's an affine transform involved
        primal_coef[:] = final_smooth.grad + idx / (idx + 3.) * (final_smooth.grad - primal_coef)
        if idx == max_iters-1:
            warnings.warn('problem not solved after %d iterations' % max_iters)
    
    return primal_coef, dual_coef



import warnings
from copy import copy

import numpy as np
from scipy import sparse
from scipy.stats import norm as normal_dbn

from . import smooth_atom, affine_smooth
from ..affine import (astransform, 
                      linear_transform, 
                      affine_transform,
                      scaler)
from ..identity_quadratic import identity_quadratic
from ..atoms.seminorms import l1norm
from .cox import cox_loglike
from .binary import (logistic_loglike,
                     probit_loglike,
                     cloglog_loglike,
                     huber_svm)

class glm(smooth_atom):

    """
    A general linear model, usually a log-likelihood
    for response $Y$ whose mean is modelled through
    $X\beta$. 

    Usual examples are Gaussian (least squares regression),
    logistic regression and Poisson log-linear regression.
    
    Huber loss is also implemented as an example.

    """

    def __init__(self, 
                 X, 
                 Y, 
                 loss, 
                 quadratic=None, 
                 initial=None,
                 offset=None,
                 case_weights=None):

        """

        Parameters
        ----------

        X : ndarray
            The design matrix.

        Y : ndarray
            The response.

        loss : `regreg.smooth.smooth_atom`
            The loss function that takes arguments the same
            size as `Y`. So, for Gaussian regression 
            the loss is just the map $\mu \mapsto \|\mu - Y\|^2_2/2$.

        quadratic : `regreg.identity_quadratic.identity_quadratic`
            Optional quadratic part added to objective.

        initial : ndarray
            An initial guess at the minimizer.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        case_weights : ndarray
            Non-negative case weights

        """

        self.saturated_loss = loss
        self.data = X, Y
        self.affine_atom = affine_smooth(loss, X)
        if case_weights is None:
            case_weights = np.ones(X.shape[0])
        self.case_weights = case_weights
        
        if self.case_weights is not None:
            if not np.all(case_weights >= 0):
                raise ValueError('case_weights should be non-negative')
            self.case_weights = np.asarray(case_weights)
            if self.case_weights.shape != loss.shape[:1]:
                raise ValueError('case_weights should have same shape as loss.shape[:1]')

        smooth_atom.__init__(self,
                             X.shape[1],
                             coef=1.,
                             offset=offset,
                             quadratic=quadratic,
                             initial=initial)

    def smooth_objective(self, 
                         beta, 
                         mode='func', 
                         check_feasibility=False):
        """

        Parameters
        ----------

        beta : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `beta` is not
            in the domain.

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `beta`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """
        beta = self.apply_offset(beta)
        linear_pred = self.affine_atom.affine_transform.dot(beta)
        value = self.saturated_loss.smooth_objective(linear_pred, 
                                                     mode=mode, 
                                                     check_feasibility=check_feasibility,
                                                     case_weights=self.case_weights)

        if mode == 'func':
            return self.scale(value)
        elif mode == 'grad':
            return self.scale(self.affine_atom.affine_transform.adjoint_map(value))
        else:
            return self.scale(value[0]), self.scale(self.affine_atom.affine_transform.adjoint_map(value[1]))

    def get_data(self):
        return self._X, self.saturated_loss.data

    def set_data(self, data):
        X, Y = data
        self._transform = astransform(X)
        self._X = X
        self._is_transform = id(self._X) == id(self._transform) # i.e. astransform was a nullop
        self.saturated_loss.data = Y

    data = property(get_data, set_data, doc="Data for the general linear model.")

    def linear_predictor(self, beta):
        """

        Compute $X\beta$.

        Parameters
        ----------

        beta : ndarray
            Parameters.

        Returns
        -------

        linpred : ndarray

        """
        # can have an `R`-type offset by using affine_map here
        return self._transform.linear_map(beta)

    def objective(self, beta):
        """
        Compute the loss $\ell(X\beta)$.

        Parameters
        ----------

        beta : ndarray
            Parameters.

        Returns
        -------

        objective : float
            Value of the loss at $\beta$.

        """
        return self.smooth_objective(beta, 'func')

    def gradient(self, beta):
        """

        Compute the gradient of the loss $ \nabla \ell(X\beta)$.

        Parameters
        ----------

        beta : ndarray
            Parameters.

        Returns
        -------

        grad : ndarray
            Gradient of the loss at $\beta$.
        """

        return self.smooth_objective(beta, 'grad')

    def hessian(self, beta):
        """

        Compute the Hessian of the loss $ \nabla^2 \ell(X\beta)$.


        Parameters
        ----------

        beta : ndarray
            Parameters.

        Returns
        -------

        hess : ndarray
            Hessian of the loss at $\beta$, if defined.

        """

        linpred = self.linear_predictor(beta)
        X = self.data[0]
        if self._is_transform:
            raise ValueError('refusing to form Hessian for arbitrary affine_transform, use an ndarray or scipy.sparse')

        if not hasattr(self.saturated_loss, 'hessian'):
            if not hasattr(self.saturated_loss, 'hessian_mult'):
                raise ValueError('loss has no hessian or hessian_mult method')
            right_mult = np.zeros(X.shape)
            for j in range(X.shape[1]):
                right_mult[:,j] = self.saturated_loss.hessian_mult(linpred, 
                                                                   X[:,j], 
                                                                   case_weights=self.case_weights)
        else:
            W = self.saturated_loss.hessian(linpred, 
                                            case_weights=self.case_weights)
            right_mult = W[:,None] * X
        if not sparse.issparse(X): # assuming it is an ndarray
            return X.T.dot(right_mult)
        else:
            return X.T * right_mult

    def latexify(self, var=None, idx=''):
        return self.affine_atom.latexify(var=var, idx=idx)

    def __copy__(self):
        klass = self.__class__
        X, Y = self.data
        return klass(copy(X), 
                     copy(Y), 
                     copy(self.saturated_loss), 
                     quadratic=copy(self.quadratic),
                     initial=copy(self.coefs),
                     offset=copy(self.offset),
                     case_weights=self.case_weights.copy())

    def subsample(self, idx):
        """
        Create a loss using a subsample of the data.
        Makes a copy of the loss and 
        multiplies case_weights by the indicator for
        `idx`.

        Parameters
        ----------

        idx : index
            Indices of np.arange(n) to keep.

        Returns
        -------

        subsample_loss : `glm`
            Loss after discarding all
            cases not in `idx.

        """

        subsample_loss = copy(self)
        n = subsample_loss.saturated_loss.shape[0]

        idx_bool = np.zeros(n, np.bool)
        idx_bool[idx] = 1

        subsample_loss.case_weights *= idx_bool

        return subsample_loss
        
    @classmethod
    def gaussian(klass,
                 X, 
                 response,
                 case_weights=None,
                 coef=1., 
                 saturated_offset=None,
                 offset=None,
                 quadratic=None, 
                 initial=None):
        """
        Create a loss for a Gaussian regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        response : ndarray
            Response vector.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in saturated space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = gaussian_loglike(response.shape,
                                response,
                                coef=coef,
                                offset=saturated_offset)

        return klass(X, 
                     response, 
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def logistic(klass, 
                 X, 
                 successes, 
                 trials=None,
                 case_weights=None,
                 coef=1., 
                 offset=None,
                 saturated_offset=None,
                 quadratic=None, 
                 initial=None):
        """
        Create a loss for a logistic regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = logistic_loglike(successes.shape,
                                successes,
                                coef=coef,
                                offset=saturated_offset,
                                trials=trials)

        return klass(X, 
                     (successes, loss.trials),
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def probit(klass, 
               X, 
               successes, 
               trials=None,
               case_weights=None,
               coef=1., 
               offset=None,
               saturated_offset=None,
               quadratic=None, 
               initial=None):
        """
        Create a loss for a probit regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = probit_loglike(successes.shape,
                              successes,
                              coef=coef,
                              offset=saturated_offset,
                              trials=trials)

        return klass(X, 
                     (successes, loss.trials),
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def cloglog(klass, 
                X, 
                successes, 
                trials=None,
                case_weights=None,
                coef=1., 
                offset=None,
                saturated_offset=None,
                quadratic=None, 
                initial=None):
        """
        Create a loss for a cloglog regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = cloglog_loglike(successes.shape,
                               successes,
                               coef=coef,
                               offset=saturated_offset,
                               trials=trials)

        return klass(X, 
                     (successes, loss.trials),
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def poisson(klass,
                X, 
                counts,
                case_weights=None,
                coef=1., 
                saturated_offset=None,
                offset=None,
                quadratic=None, 
                initial=None):
        """
        Create a loss for a Poisson regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        counts : ndarray
            Response vector. Should be non-negative integers.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = poisson_loglike(counts.shape,
                               counts,
                               offset=saturated_offset,
                               coef=coef)

        return klass(X, counts, loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def huber(klass,
              X, 
              response,
              smoothing_parameter,
              case_weights=None,
              coef=1., 
              saturated_offset=None,
              offset=None,
              quadratic=None, 
              initial=None):
        """
        Create a loss for a regression model using
        Huber loss.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        response : ndarray
            Response vector. 

        smoothing_parameter : float
            Smoothing parameter for Huber loss.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = huber_loss(response.shape,
                          response,
                          smoothing_parameter,
                          offset=saturated_offset,
                          coef=coef)

        return klass(X, 
                     response, 
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def huber_svm(klass,
                  X, 
                  successes,
                  smoothing_parameter,
                  case_weights=None,
                  coef=1., 
                  offset=None,
                  saturated_offset=None,
                  quadratic=None, 
                  initial=None):
        """
        Create a loss for a binary regression model using
        Huber SVM loss.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        successes : ndarray
            Response vector. 

        smoothing_parameter : float
            Smoothing parameter for Huber loss.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = huber_svm(successes.shape,
                         successes,
                         smoothing_parameter,
                         offset=saturated_offset,
                         coef=coef)

        return klass(X, 
                     successes,
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def cox(klass, 
            X, 
            event_times,
            censoring,
            case_weights=None,
            coef=1., 
            saturated_offset=None,
            offset=None,
            quadratic=None, 
            initial=None):
        """
        Create a loss for a logistic regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        event_times : ndarray
            Observed times for Cox proportional hazard model.

        censoring : ndarray 
            Censoring indicator for Cox proportional hazard model
            - 1 indicates observation is a failure, 0 a censored observation.

        case_weights : ndarray
            Non-negative case weights

        coef : float
            Scaling to be put in front of loss.

        saturated_offset : ndarray (optional)
            Offset to be applied in saturated parameter space before 
            evaluating loss.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        glm_obj : `regreg.glm.glm`
            General linear model loss.

        """

        loss = cox_loglike(event_times.shape,
                           event_times,
                           censoring,
                           offset=saturated_offset,
                           coef=coef)

        return klass(X, 
                     np.array([event_times, censoring]).T,
                     loss,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)


class gaussian_loglike(smooth_atom):

    """
    The Gaussian loss for observations $y$:

    .. math::
    
       \mu \mapsto \frac{1}{2} \|y-\mu\|^2_2

    """

    objective_template = r"""\ell^{\text{Gauss}}\left(%(var)s\right)"""

    def __init__(self, 
                 shape,
                 response, 
                 coef=1., 
                 offset=None,
                 quadratic=None,
                 initial=None,
                 case_weights=None):

        smooth_atom.__init__(self,
                             shape,
                             offset=offset,
                             quadratic=quadratic,
                             initial=initial,
                             coef=coef)

        if sparse.issparse(response):
            self.response = response.toarray().flatten()
        else:
            self.response = np.asarray(response)

        if case_weights is not None:
            if not np.all(case_weights >= 0):
                raise ValueError('case_weights should be non-negative')
            self.case_weights = np.asarray(case_weights)
            if self.case_weights.shape != self.response.shape:
                raise ValueError('case_weights should have same shape as response')
        else:
            self.case_weights = None

    def smooth_objective(self, 
                         natural_param, 
                         mode='both', 
                         check_feasibility=False,
                         case_weights=None):
        """

        Evaluate the smooth objective, computing its value, gradient or both.

        Parameters
        ----------

        natural_param : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `natural_param` is not
            in the domain.

        case_weights : ndarray
            Non-negative case weights

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `natural_param`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """
        
        if case_weights is None:
            case_weights = np.ones_like(natural_param)
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights

        natural_param = self.apply_offset(natural_param)
        resid = natural_param - self.response 

        if mode == 'both':
            f, g = self.scale(np.sum(cw*resid**2)) / 2., self.scale(cw*resid)
            return f, g
        elif mode == 'grad':
            return self.scale(cw*resid) 
        elif mode == 'func':
            return self.scale(np.sum(cw*resid**2)) / 2.
        else:
            raise ValueError("mode incorrectly specified")
            
    # Begin loss API

    def hessian(self, 
                natural_param, 
                case_weights=None):
        """
        Hessian of the loss.

        Parameters
        ----------

        natural_param : ndarray
            Parameters where Hessian will be evaluated.

        Returns
        -------

        hess : ndarray
            A 1D-array representing the diagonal of the Hessian
            evaluated at `natural_param`.

        case_weights : ndarray
            Non-negative case weights

        """
        if case_weights is None:
            case_weights = np.ones_like(natural_param)
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights
        return self.scale(np.ones_like(natural_param) * cw)

    def get_data(self):
        return self.response

    def set_data(self, data):
        self.response = data

    data = property(get_data, set_data)

    def __copy__(self):
        return gaussian_loglike(self.shape,
                                copy(self.response),
                                coef=self.coef, 
                                offset=copy(self.offset),
                                quadratic=copy(self.quadratic),
                                initial=copy(self.coefs),
                                case_weights=copy(self.case_weights))

    def subsample(self, case_idx):
        """
        Create a saturated loss using a subsample of the data.
        Makes a copy of the loss and 
        multiplies case_weights by the indicator for
        `idx`.

        Parameters
        ----------

        idx : index
            Indices of np.arange(n) to keep.

        Returns
        -------

        subsample_loss : `smooth_atom`
            Loss after discarding all
            cases not in `idx.

        """
        loss_cp = copy(self)
        if loss_cp.case_weights is None:
            case_weights = loss_cp.case_weights = np.ones(self.shape[0])
        else:
            case_weights = loss_cp.case_weights

        idx_bool = np.zeros_like(case_weights, np.bool)
        idx_bool[case_idx] = 1

        case_weights *= idx_bool
        return loss_cp

    # End loss API

    def mean_function(self, eta):
        return eta

class poisson_loglike(smooth_atom):

    """
    A class for combining the Poisson log-likelihood with a general seminorm
    """

    objective_template = r"""\ell^{\text{Pois}}\left(%(var)s\right)"""

    def __init__(self, 
                 shape, 
                 counts, 
                 coef=1., 
                 offset=None,
                 quadratic=None,
                 initial=None,
                 case_weights=None):

        smooth_atom.__init__(self,
                             shape,
                             offset=offset,
                             quadratic=quadratic,
                             initial=initial,
                             coef=coef)

        if sparse.issparse(counts):
            #Convert sparse success vector to an array
            self.counts = counts.toarray().flatten()
        else:
            self.counts = counts

        if not np.allclose(np.round(self.counts),self.counts):
            raise ValueError("Counts vector is not integer valued")
        if np.min(self.counts) < 0:
            raise ValueError("Counts vector is not non-negative")

        saturated = counts
        loss_terms = - coef * ((counts - 1) * np.log(counts))
        loss_terms[counts == 0] = 0

        loss_constant = - coef * loss_terms.sum()

        devq = identity_quadratic(0,0,0,-loss_constant)
        self.quadratic += devq

        if case_weights is not None:
            if not np.all(case_weights >= 0):
                raise ValueError('case_weights should be non-negative')
            self.case_weights = np.asarray(case_weights)
        else:
            self.case_weights = None

    def smooth_objective(self, 
                         natural_param, 
                         mode='both', 
                         check_feasibility=False,
                         case_weights=None):
        """

        Evaluate the smooth objective, computing its value, gradient or both.

        Parameters
        ----------

        natural_param : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `natural_param` is not
            in the domain.

        case_weights : ndarray
            Non-negative case weights

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `natural_param`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """

        if case_weights is None:
            case_weights = np.ones_like(natural_param)
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights

        x = natural_param # shorthand

        x = self.apply_offset(x)
        exp_x = np.exp(x)
        
        if mode == 'both':
            f, g = - self.scale(-np.sum(cw * exp_x) + np.dot(cw * self.counts,x)), - self.scale(cw * (self.counts - exp_x))
            return f, g
        elif mode == 'grad':
            f, g = None, - self.scale(cw * (self.counts - exp_x))
            return g
        elif mode == 'func':
            f, g =  - self.scale(-np.sum(cw * exp_x) + np.dot(cw * self.counts,x)), None
            return f
        else:
            raise ValueError("mode incorrectly specified")

    # Begin loss API

    def hessian(self, natural_param, case_weights=None):
        """
        Hessian of the loss.

        Parameters
        ----------

        natural_param : ndarray
            Parameters where Hessian will be evaluated.

        case_weights : ndarray
            Non-negative case weights

        Returns
        -------

        hess : ndarray
            A 1D-array representing the diagonal of the Hessian
            evaluated at `natural_param`.
        """
        x = natural_param # shorthand

        if case_weights is None:
            case_weights = np.ones_like(natural_param)
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights

        return self.scale(cw * np.exp(x))
            
    def get_data(self):
        return self.counts

    def set_data(self, data):
        self.counts = data

    data = property(get_data, set_data)

    def __copy__(self):
        counts = self.data
        return poisson_loglike(self.shape,
                               copy(counts),
                               coef=self.coef,
                               offset=copy(self.offset),
                               quadratic=copy(self.quadratic),
                               initial=copy(self.coefs),
                               case_weights=copy(self.case_weights))

    def subsample(self, case_idx):
        """
        Create a saturated loss using a subsample of the data.
        Makes a copy of the loss and 
        multiplies case_weights by the indicator for
        `idx`.

        Parameters
        ----------

        idx : index
            Indices of np.arange(n) to keep.

        Returns
        -------

        subsample_loss : `smooth_atom`
            Loss after discarding all
            cases not in `idx.

        """
        loss_cp = copy(self)
        if loss_cp.case_weights is None:
            case_weights = loss_cp.case_weights = np.ones(self.shape[0])
        else:
            case_weights = loss_cp.case_weights

        idx_bool = np.zeros_like(case_weights, np.bool)
        idx_bool[case_idx] = 1

        case_weights *= idx_bool
        return loss_cp

    # End loss API

    def mean_function(self, eta):
        return np.exp(eta)

class huber_loss(smooth_atom):

    objective_template = r"""\ell^{\text{Huber}}\left(%(var)s\right)"""

    def __init__(self, 
                 shape, 
                 response, 
                 smoothing_parameter,
                 coef=1., 
                 offset=None,
                 quadratic=None,
                 initial=None,
                 case_weights=None):

        smooth_atom.__init__(self,
                             shape,
                             offset=offset,
                             quadratic=quadratic,
                             initial=initial,
                             coef=coef)

        self.smoothing_parameter = smoothing_parameter
        atom = l1norm(shape, lagrange=1.)
        Q = identity_quadratic(smoothing_parameter, 0, 0, 0)
        self.smoothed_atom = atom.smoothed(Q)
                                         
        if sparse.issparse(response):
            self.response = response.toarray().flatten()
        else:
            self.response = np.asarray(response)

        if case_weights is not None:
            if not np.all(case_weights >= 0):
                raise ValueError('case_weights should be non-negative')
            self.case_weights = np.asarray(case_weights)
            if self.case_weights.shape != self.response.shape:
                raise ValueError('case_weights should have same shape as response')
        else:
            self.case_weights = None

    def smooth_objective(self, 
                         param, 
                         mode='both', 
                         check_feasibility=False,
                         case_weights=None):
        """

        Evaluate the smooth objective, computing its value, gradient or both.

        Parameters
        ----------

        param : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `param` is not
            in the domain.

        case_weights : ndarray
            Non-negative case weights

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `param`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """
        
        if case_weights is None:
            case_weights = np.ones_like(param)
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights

        param = self.apply_offset(param)
        resid = param - self.response 

        f, g = _huber_loss(resid, smoothing_parameter=self.smoothing_parameter)

        if mode == 'func':
            return self.scale((f * cw).sum())
        elif mode == 'grad':
            return self.scale(g * cw)
        elif mode == 'both':
            return self.scale((f * cw).sum()), self.scale(g * cw)
        else:
            raise ValueError("mode incorrectly specified")

    # Begin loss API

    def hessian(self, param, case_weights=None):
        """
        Hessian of the loss.

        Parameters
        ----------

        param : ndarray
            Parameters where Hessian will be evaluated.

        case_weights : ndarray
            Non-negative case weights

        Returns
        -------

        hess : ndarray
            A 1D-array representing the diagonal of the Hessian
            evaluated at `natural_param`.
        """

        # it is piecwise C^2 though... maybe use this?
        raise NotImplementedError('Huber loss is not twice differentiable')

    def get_data(self):
        return self.response

    def set_data(self, data):
        self.response = data

    data = property(get_data, set_data)

    def subsample(self, case_idx):
        """
        Create a saturated loss using a subsample of the data.
        Makes a copy of the loss and 
        multiplies case_weights by the indicator for
        `idx`.

        Parameters
        ----------

        idx : index
            Indices of np.arange(n) to keep.

        Returns
        -------

        subsample_loss : `smooth_atom`
            Loss after discarding all
            cases not in `idx.

        """
        loss_cp = copy(self)
        if loss_cp.case_weights is None:
            case_weights = loss_cp.case_weights = np.ones(self.shape[0])
        else:
            case_weights = loss_cp.case_weights

        idx_bool = np.zeros_like(case_weights, np.bool)
        idx_bool[case_idx] = 1

        case_weights *= idx_bool
        return loss_cp

    def __copy__(self):
        response = self.data
        return huber_loss(self.shape,
                          copy(response),
                          self.smoothing_parameter,
                          coef=self.coef,
                          offset=copy(self.offset),
                          quadratic=copy(self.quadratic),
                          initial=copy(self.coefs))

    # End loss API

def _huber_loss(arg, smoothing_parameter):
    # returns vector whose sum is total loss as well as gradient vector
    eps = smoothing_parameter
    proj_arg = np.sign(arg) * np.minimum(np.abs(arg) / eps, 1) # the maximizer is the gradient
                                                               # by convex conjugacy
    return arg * proj_arg - eps * proj_arg**2 / 2, proj_arg 

class stacked_loglike(smooth_atom):

    """
    A class for stacking `K=len(losses)` saturated losses with common
    shapes (roughly speaking) summed over losses.

    Roughly speaking a model of `K` independent measurements per individual.
    """

    objective_template = r"""\ell^{\text{stack}}\left(%(var)s\right)"""

    def __init__(self, 
                 losses,
                 coef=1., 
                 offset=None,
                 quadratic=None,
                 initial=None,
                 case_weights=None):

        shape = (np.sum([l.shape[0] for l in losses]),)
        smooth_atom.__init__(self,
                             shape,
                             offset=offset,
                             quadratic=quadratic,
                             initial=initial,
                             coef=coef)

        responses = [l.data for l in losses]
        shapes = [r.shape for r in responses]
        dims = np.array([len(s) for s in shapes])
        if np.all(dims == 1):
            self.data = np.hstack(responses)
        elif np.all(dims == 2):
            self.data = np.vstack(responses)
        else:
            raise ValueError('expecting either 1 or dimensional data for saturated losses')

        self._slices = []
        idx = 0
        for l in losses:
            self._slices.append(slice(idx, idx + l.shape[0], 1))
            idx += l.shape[0]

        self._losses = losses
        self._gradient = np.zeros(self.shape)

        if case_weights is not None:
            if not np.all(case_weights >= 0):
                raise ValueError('case_weights should be non-negative')
            self.case_weights = np.asarray(case_weights)
            if self.case_weights.shape != self.shape[:1]:
                raise ValueError('case_weights should have same shape as response')
        else:
            self.case_weights = None

    def smooth_objective(self, 
                         natural_param, 
                         mode='both', 
                         check_feasibility=False,
                         case_weights=None):
        """

        Evaluate the smooth objective, computing its value, gradient or both.

        Parameters
        ----------

        natural_param : ndarray
            The current parameter values.

        mode : str
            One of ['func', 'grad', 'both']. 

        check_feasibility : bool
            If True, return `np.inf` when
            point is not feasible, i.e. when `natural_param` is not
            in the domain.

        Returns
        -------

        If `mode` is 'func' returns just the objective value 
        at `natural_param`, else if `mode` is 'grad' returns the gradient
        else returns both.
        """
        
        if case_weights is None:
            case_weights = np.ones(natural_param.shape[:1])
        cw = case_weights
        if self.case_weights is not None:
            cw *= self.case_weights

        linpred = natural_param # shorthand

        linpred = self.apply_offset(linpred)
        if mode == 'grad':
            for d, slice in enumerate(self._slices):
                self._gradient[slice] = self._losses[d].smooth_objective(linpred[slice], 'grad')
            return self.scale(self._gradient)
        elif mode == 'func':
            value = 0
            for d, slice in enumerate(self._slices):
                value += self._losses[d].smooth_objective(linpred[slice], 'func')
            return self.scale(value)
        elif mode == 'both':
            value = 0
            for d, slice in enumerate(self._slices):
                f, g = self._losses[d].smooth_objective(linpred[slice], 'both')
                self._gradient[slice] = g
                value += f
            return self.scale(value), self.scale(self._gradient)
        else:
            raise ValueError("mode incorrectly specified")

    def get_data(self):
        return self._data

    def set_data(self, data):
        self._data = data

    data = property(get_data, set_data)

    def __copy__(self):
        return stacked_loglike(copy(self._losses),
                               coef=self.coef,
                               offset=copy(self.offset),
                               quadratic=copy(self.quadratic),
                               initial=copy(self.coefs),
                               case_weights=copy(self.case_weights))

    def subsample(self, case_idx):
        """
        Create a saturated loss using a subsample of the data.
        Makes a copy of the loss and 
        multiplies case_weights by the indicator for
        `idx`.

        Parameters
        ----------

        idx : index
            Indices of np.arange(n) to keep.

        Returns
        -------

        subsample_loss : `smooth_atom`
            Loss after discarding all
            cases not in `idx.

        """
        loss_cp = copy(self)
        if loss_cp.case_weights is None:
            case_weights = loss_cp.case_weights = np.ones(self.shape[0])
        else:
            case_weights = loss_cp.case_weights

        idx_bool = np.zeros_like(case_weights, np.bool)
        idx_bool[case_idx] = 1

        case_weights *= idx_bool
        return loss_cp

    @classmethod
    def gaussian(klass,
                 responses,
                 case_weights=None,
                 coef=1., 
                 offset=None,
                 quadratic=None, 
                 initial=None):
        """
        Create a loss for a Gaussian regression model.

        Parameters
        ----------

        responses : ndarray
            Response vectors.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [gaussian_loglike(response.shape,
                                   response,
                                   coef=coef)
                  for response in responses]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def logistic(klass, 
                 successes, 
                 trials=None,
                 case_weights=None,
                 coef=1., 
                 offset=None,
                 quadratic=None, 
                 initial=None):
        """
        Create a loss for a logistic regression model.

        Parameters
        ----------

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [logistic_loglike(successes[i],
                                   trials[i],
                                   coef=coef)
                  for i in range(len(successes))]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def probit(klass, 
               successes, 
               trials=None,
               case_weights=None,
               coef=1., 
               offset=None,
               quadratic=None, 
               initial=None):
        """
        Create a loss for a probit regression model.

        Parameters
        ----------

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [probit_loglike(successes[i],
                                 trials[i],
                                 coef=coef)
                  for i in range(len(successes))]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def cloglog(klass, 
                successes, 
                trials=None,
                case_weights=None,
                coef=1., 
                offset=None,
                quadratic=None, 
                initial=None):
        """
        Create a loss for a logistic regression model.

        Parameters
        ----------

        successes : ndarray
            Responses (should be non-negative integers).

        trials : ndarray (optional)
            Number of trials for each success. If `None`,
            defaults to `np.ones_like(successes)`.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [cloglog_loglike(successes[i],
                                  trials[i],
                                  coef=coef)
                  for i in range(len(successes))]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def poisson(klass,
                counts,
                case_weights=None,
                coef=1., 
                offset=None,
                quadratic=None, 
                initial=None):
        """
        Create a loss for a Poisson regression model.

        Parameters
        ----------

        counts : ndarray
            Response vector. Should be non-negative integers.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [logistic_loglike(successes[i],
                                   trials[i],
                                   coef=coef)
                  for i in range(len(successes))]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def huber(klass,
              X, 
              responses,
              smoothing_parameter,
              case_weights=None,
              coef=1., 
              offset=None,
              quadratic=None, 
              initial=None):
        """
        Create a loss for a regression model using
        Huber loss.

        Parameters
        ----------

        response : ndarray
            Response vector. 

        smoothing_parameter : float
            Smoothing parameter for Huber loss.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [huber_loss(response.shape,
                             response,
                             smoothing_parameter) 
                  for response in responses]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def huber_svm(klass,
                  X, 
                  successes,
                  smoothing_parameter,
                  case_weights=None,
                  coef=1., 
                  offset=None,
                  quadratic=None, 
                  initial=None):
        """
        Create a loss for a regression model using
        Huber loss.

        Parameters
        ----------

        response : ndarray
            Response vector. 

        smoothing_parameter : float
            Smoothing parameter for Huber loss.

        case_weights : ndarray
            Non-negative case weights

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [huber_svm(response.shape,
                            response,
                            smoothing_parameter) 
                  for response in successes]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)

    @classmethod
    def cox(klass, 
            X, 
            event_times,
            censoring,
            coef=1., 
            offset=None,
            quadratic=None, 
            initial=None,
            case_weights=None):
        """
        Create a loss for a Cox regression model.

        Parameters
        ----------

        X : [ndarray, `regreg.affine.affine_transform`]
            Design matrix

        event_times : ndarray
            Observed times for Cox proportional hazard model.

        censoring : ndarray 
            Censoring indicator for Cox proportional hazard model
            - 1 indicates observation is a failure, 0 a censored observation.

        offset : ndarray (optional)
            Offset to be applied in parameter space before 
            evaluating loss.

        quadratic : `regreg.identity_quadratic.identity_quadratic` (optional)
            Optional quadratic to be added to objective.

        initial : ndarray
            Initial guess at coefficients.
           
        Returns
        -------

        mglm_obj : `regreg.mglm.mglm`
            General linear model loss.

        """

        losses = [cox_loglike(event_times[i].shape,
                              event_times[i],
                              censoring[i],
                              coef=coef)
                  for i in range(len(event_times))]

        return klass(losses,
                     offset=offset,
                     quadratic=quadratic,
                     initial=initial,
                     case_weights=case_weights)


# Deprecated

def logistic_loss(X, Y, trials=None, coef=1.):
    '''
    Construct a logistic loss function for successes Y and
    affine transform X.

    Parameters
    ----------

    X : [affine_transform, ndarray]
        Design matrix

    Y : ndarray

    '''
    warnings.warn('"logistic_loss" is deprecated use "regreg.smooth.glm.logistic" instead')
    return glm.logistic(X, 
                        Y,
                        trials=trials,
                        coef=coef)




import numpy as np
try:
    from sklearn.model_selection import cross_validate
    from sklearn.model_selection import GridSearchCV
    from ..api import (gaussian,
                       gaussian_lagrange)
    have_sklearn = True
except ImportError:
    have_sklearn = False

from ...api import l1norm, group_lasso
from ...tests.decorators import set_seed_for_test

@np.testing.dec.skipif(not have_sklearn)
@set_seed_for_test()
def test_regression_gaussian():

    n, p = 100, 20
    X = np.random.standard_normal((n, p))
    y = np.random.standard_normal(n)

    atom = l1norm
    atom_args = {'shape':p, 'lagrange':2*np.sqrt(n)}

    gaussian_lasso = gaussian(atom, atom_args)
    gaussian_lasso.fit(X, y)
    print(cross_validate(gaussian_lasso, X, y, cv=10))

    # grid search
    params = {'atom_params':[{'shape':p,
                              'lagrange': alpha * 2 * np.sqrt(n)} for alpha in [0.5, 1, 1.5]]}
    lasso_cv = GridSearchCV(gaussian_lasso, params, cv=3)
    lasso_cv.fit(X, y)

    gaussian_lasso_offset = gaussian(atom, atom_args,
                                             offset=True)
    y_offset = np.array([y, np.random.standard_normal(n)]).T
    gaussian_lasso_offset.fit(X, y_offset)

    gaussian_lasso_weights = gaussian(atom, atom_args,
                                      case_weights=True,
                                      score_method='mean_deviance')

    y_weights = np.array([y, np.ones_like(y)]).T
    gaussian_lasso_weights.fit(X, y_weights)

    np.testing.assert_allclose(gaussian_lasso_weights._coefs,
                               gaussian_lasso._coefs)

    GridSearchCV(gaussian_lasso_offset, params, cv=3).fit(X, y_offset)
    GridSearchCV(gaussian_lasso_weights, params, cv=3).fit(X, y_weights)

    gaussian_lasso_both = gaussian(atom, atom_args,
                                   offset=True,
                                   case_weights=True,
                                   score_method='R2')
    y_both = np.array([y, np.ones_like(y), y_offset[:,1]]).T
    GridSearchCV(gaussian_lasso_both, params, cv=3).fit(X, y_both)

    def atom_constructor(null_grad, **args):
        return l1norm(**args)
    gaussian_lasso_enet = gaussian_lagrange(atom_constructor,
                                            atom_args,
                                            offset=True,
                                            enet_alpha=0.5,
                                            case_weights=True,
                                            score_method='R2')
    y_both = np.array([y, np.ones_like(y), y_offset[:,1]]).T
    GridSearchCV(gaussian_lasso_enet, params, cv=3).fit(X, y_both)

@np.testing.dec.skipif(not have_sklearn)
@set_seed_for_test()
def test_regression_group_lasso():

    n, p = 100, 20
    X = np.random.standard_normal((n, p))
    y = np.random.standard_normal(n)

    atom = group_lasso
    atom_args = {'groups':[0]*10+[1]*10, 'lagrange':2*np.sqrt(n)}

    gaussian_lasso = gaussian(atom, atom_args)
    gaussian_lasso.fit(X, y)
    print(cross_validate(gaussian_lasso, X, y, cv=10))

    # grid search
    params = {'atom_params':[{'groups':[0]*10+[1]*10,
                              'lagrange': alpha * 2 * np.sqrt(n)} for alpha in [0.5, 1, 1.5]]}
    lasso_cv = GridSearchCV(gaussian_lasso, params, cv=3)
    lasso_cv.fit(X, y)

    gaussian_lasso_offset = gaussian(atom, atom_args,
                                             offset=True)
    y_offset = np.array([y, np.random.standard_normal(n)]).T
    gaussian_lasso_offset.fit(X, y_offset)

    gaussian_lasso_weights = gaussian(atom, atom_args,
                                      case_weights=True,
                                      score_method='mean_deviance')

    y_weights = np.array([y, np.ones_like(y)]).T
    gaussian_lasso_weights.fit(X, y_weights)

    np.testing.assert_allclose(gaussian_lasso_weights._coefs,
                               gaussian_lasso._coefs)

    GridSearchCV(gaussian_lasso_offset, params, cv=3).fit(X, y_offset)
    GridSearchCV(gaussian_lasso_weights, params, cv=3).fit(X, y_weights)

    gaussian_lasso_both = gaussian(atom, atom_args,
                                   offset=True,
                                   case_weights=True,
                                   score_method='R2')
    y_both = np.array([y, np.ones_like(y), y_offset[:,1]]).T
    GridSearchCV(gaussian_lasso_both, params, cv=3).fit(X, y_both)

    def atom_constructor(null_grad, **args):
        return group_lasso(**args)
    gaussian_lasso_enet = gaussian_lagrange(atom_constructor,
                                            atom_args,
                                            offset=True,
                                            enet_alpha=0.5,
                                            case_weights=True,
                                            score_method='R2')
    y_both = np.array([y, np.ones_like(y), y_offset[:,1]]).T
    GridSearchCV(gaussian_lasso_enet, params, cv=3).fit(X, y_both)

@np.testing.dec.skipif(not have_sklearn)
@set_seed_for_test()
def test_regression_gaussian_bound():

    n, p = 100, 20
    X = np.random.standard_normal((n, p))
    y = np.random.standard_normal(n)

    atom = l1norm
    atom_args = {'shape':p, 'bound':3}

    gaussian_lasso = gaussian(atom,
                              atom_args)

    gaussian_lasso.fit(X, y)
    print(cross_validate(gaussian_lasso, X, y, cv=10))

    # grid search
    params = {'atom_params':[{'shape':p,
                              'bound': b,
                              } for b in [3, 4, 5]]}
    lasso_cv = GridSearchCV(gaussian_lasso, params, cv=3)
    lasso_cv.fit(X, y)

    gaussian_lasso_offset = gaussian(atom,
                                     atom_args,
                                     offset=True)
    y_offset = np.array([y, np.random.standard_normal(n)]).T
    gaussian_lasso_offset.fit(X, y_offset)

    gaussian_lasso_weights = gaussian(atom,
                                      atom_args,
                                      case_weights=True,
                                      score_method='mean_deviance')

    y_weights = np.array([y, np.ones_like(y)]).T
    gaussian_lasso_weights.fit(X, y_weights)

    np.testing.assert_allclose(gaussian_lasso_weights._coefs,
                               gaussian_lasso._coefs)

    GridSearchCV(gaussian_lasso_offset, params, cv=3).fit(X, y_offset)
    GridSearchCV(gaussian_lasso_weights, params, cv=3).fit(X, y_weights)

    gaussian_lasso_both = gaussian(atom,
                                   atom_args,
                                   offset=True,
                                   case_weights=True,
                                   score_method='R2')
    y_both = np.array([y, np.ones_like(y), y_offset[:,1]]).T
    GridSearchCV(gaussian_lasso_both, params, cv=3).fit(X, y_both)







import numpy as np
import nose.tools as nt

from regreg.smooth.losses import huberized_svm

def test_huberized_svm():

    Y = np.random.binomial(1,0.5,size=(10,))
    X = np.random.standard_normal((10,5))

    L = huberized_svm(X, Y, 0.01)
    L.smooth_objective(np.zeros(L.shape), 'both')

    L.latexify()



import itertools
import numpy as np
import nose.tools as nt

import regreg.api as rr
from ..sparse_group_block import (sparse_group_block,
                                             sparse_group_block_dual,
                                             _inside_set,
                                             _gauge_function_dual)
from ..sparse_group_lasso import sparse_group_lasso
from regreg.tests.decorators import set_seed_for_test

from .test_seminorms import Solver, SolverFactory, all_close

def test_sparse_group_lasso_equivalent():
    """
    check agrees with equivalent group lasso
    """

    l1_weight = 1.
    l2_weight = 2.
    lagrange = 0.5
    pen = sparse_group_block((4, 5), l1_weight, l2_weight, lagrange=lagrange)
    pen2 = sparse_group_lasso(np.multiply.outer(np.arange(4), np.ones(5)).reshape(-1).astype(np.int),
                              l1_weight,
                              weights=dict([(j, l2_weight) for j in range(4)]),
                              lagrange=lagrange)

    Z = np.random.standard_normal(pen.shape) * 2
    np.testing.assert_allclose(pen.lagrange_prox(Z), pen2.lagrange_prox(Z.reshape(-1)).reshape(Z.shape))

    dual = pen.conjugate
    dual2 = pen2.conjugate
    np.testing.assert_allclose(Z, pen.lagrange_prox(Z) + dual.bound_prox(Z))
    np.testing.assert_allclose(dual.bound_prox(Z), dual2.bound_prox(Z.reshape(-1)).reshape(Z.shape))

def test_inside_set():
    """
    with 0 as lasso weights should be group lasso
    """
    l1_weight, l2_weight = 1, 2
    pen = sparse_group_block((4, 5),
                             l1_weight, 
                             l2_weight,
                             lagrange=0.4)
    point = np.zeros(pen.shape)
    assert _inside_set(point, pen.lagrange, l1_weight, l2_weight)

    Z = np.random.standard_normal(pen.shape) * 20
    proxZ = pen.lagrange_prox(Z)
    assert _inside_set(Z - proxZ, pen.lagrange, l1_weight, l2_weight)  # its gauge norm is 0.4, larger than 1

    assert np.fabs(_gauge_function_dual(Z - proxZ, l1_weight, l2_weight) - pen.lagrange) < 1.e-4

    pen2 = sparse_group_block((4,5),
                              l1_weight, 
                              l2_weight,
                              lagrange=1.2)
    point = np.zeros(pen2.shape)
    assert _inside_set(point, pen.lagrange, l1_weight, l2_weight) 

    Z = np.random.standard_normal(pen.shape) * 20
    proxZ = pen2.lagrange_prox(Z)
    print(Z, proxZ)
    assert np.fabs(_gauge_function_dual(Z - proxZ, l1_weight, l2_weight) - 1.2) < 1.e-4
    assert not _inside_set(Z - proxZ, 1., l1_weight, l2_weight) # its gauge norm is 1.2, larger than 1

class SparseBlockSolverFactory(SolverFactory):

    shape = (4, 5)
    l1_weight = 1.
    l2_weight = 2.
    FISTA_choices = [True]
    L_choices = [0.3]
    coef_stop_choices = [False]

    def __init__(self, klass, mode):
        self.klass = klass
        self.mode = mode

    def __iter__(self):
        for offset, FISTA, coef_stop, L, q in itertools.product(self.offset_choices,
                                                                self.FISTA_choices,
                                                                self.coef_stop_choices,
                                                                self.L_choices,
                                                                self.quadratic_choices):
            self.FISTA = FISTA
            self.coef_stop = coef_stop
            self.L = L

            if self.mode == 'lagrange':
                atom = self.klass(self.shape,
                                  self.l1_weight,
                                  self.l2_weight, lagrange=self.lagrange)
            else:
                atom = self.klass(self.shape,
                                  self.l1_weight,
                                  self.l2_weight, bound=self.bound)

            if q: 
                atom.quadratic = rr.identity_quadratic(0, 0, np.random.standard_normal(atom.shape)*0.02)

            if offset:
                atom.offset = 0.02 * np.random.standard_normal(atom.shape)

            solver = Solver(atom, interactive=self.interactive, 
                            coef_stop=coef_stop,
                            FISTA=FISTA,
                            L=L)
            yield solver

@set_seed_for_test()
@np.testing.dec.slow
def test_proximal_maps():
    for klass, mode in zip([sparse_group_block,
                            sparse_group_block_dual], 
                           ['lagrange', 'bound']):
        factory = SparseBlockSolverFactory(klass, mode)
        for solver in factory:
            for t in solver.all_tests():
                yield t





"""

If `sklearn` is importable, this module defines mixin classes that pair
regression or classification losses with an atom so that cross-validation
error can be evaluated.

"""

import numpy as np

from regreg.smooth.glm import (glm, 
                               gaussian_loglike,
                               huber_loss,
                               logistic_loglike,
                               poisson_loglike,
                               cox_loglike)

from regreg.api import (simple_problem,
                        identity_quadratic)

have_sklearn = True
try:
    from sklearn.base import (BaseEstimator, 
                              RegressorMixin,
                              ClassifierMixin)
except ImportError:
    have_sklearn = False


class base_mixin(object):

    def __init__(self, 
                 atom_constructor, 
                 atom_params, 
                 case_weights=False,
                 offset=False,
                 coef=1., 
                 quadratic=None,
                 score_method='deviance',
                 solve_args={},
                 initial=None):

        """

        Parameters
        ----------

        atom_constructor : callable

            Atom constructor that is to to be used as a regularizer.
            Final atom will be constructed as `atom_constructor(**atom_params)`.

        atom_params : dict
            Dictionary of arguments to construct atom.

        offset : bool
            While `y` include an offset column when fitting?

        case_weights : bool
            Will `y` include a case weight column when fitting?

        coef : float
            Scaling to be put in front of loss.

        score_method : str
            Which score to use as default `score`?
            One of ['deviance', 'mean_deviance', 'R2']

        solve_args : dict
            Keyword arguments passed to `simple_problem.solve`

        initial : ndarray
            Warmstart.

        Notes
        -----

        If `case_weights` or `offset` is True, then the `fit`
        method should pass a 2-d array for `y` rather than just a
        response. If only one of these is not None, then the
        quantity is assumed to be the last column.  If both are
        not None, then the 2nd to last are to be the case weights
        and the last is to be the offset.

        The `offset` argument, if not None
        is *subtracted* from the linear predictor
        before evaluating the loss -- this sign is different
        than behavior in e.g. `R` where it is *added* instead.

        """

        self.atom_constructor = atom_constructor
        self.atom_params = atom_params
        self.case_weights = case_weights
        self.coef = coef
        self.offset = offset
        self.score_method = score_method
        self.solve_args = solve_args
        self.initial = initial
        self.quadratic = quadratic

    def fit(self, X, y):
        """
        Fit a regularized regression estimator.
        
            Parameters
        ----------

        X : np.ndarray((n, p))
            Feature matrix.

        y : np.ndarray(n)
            Response vector.

        Returns
        -------

        self

        """

        self._loglike = loglike = self._loglike_factory(X, y)
        atom_ = self._construct_atom()
        problem = simple_problem(loglike, atom_)
        if self.initial is not None:
            problem.coefs[:] = self.initial
        self._coefs = problem.solve(**self.solve_args)
        return self

    def predict(self, X):
        """
        Predict new response in regression setting.

        Parameters
        ----------

        X : np.ndarray((n, p))
            Feature matrix.

        Returns
        -------

        yhat : np.ndarray(n)
            Predictions from regression model.

        """
        return X.dot(self._coefs)

    def score(self, X, y, sample_weight=None):
        '''

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True values for X.

        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.

        Returns
        -------
        score : float
        '''

        response, case_weights, offset = self._check_y_arg(y)
        predictions = X.dot(self._coefs)
        if offset is not None:
            predictions -= offset   # this is how offsets are incorporated
                                    # in regreg
        if sample_weight is not None:
            if case_weights is not None:
                case_weights *= sample_weight
        return self._saturated_score(predictions, response, case_weights=case_weights)

    def _construct_atom(self):
        return self.atom_constructor(**self.atom_params)

    def _loglike_factory(self, X, y):
        raise NotImplementedErorr

    def _check_y_arg(self, y):
        """
        Find response, case weights and 
        offset (if applicable) from
        `y` argument to fit.
        """
        response_idx, case_weights_idx, offset_idx = {
            (False, False): (None, None, None),
            (False, True): (0, None, -1),
            (True, False): (0, -1, None),
            (True, True): (0, -2, -1)}[self.case_weights,
                                     self.offset]

        if response_idx is None:
            response = y
            case_weights = offset = None
        else:
            neg_index = 0

            if case_weights_idx is not None:
                case_weights = y[:,case_weights_idx]
                neg_index -= 1
            else:
                case_weights = None

            if offset_idx is not None:
                offset = y[:,offset_idx]
                neg_index -= 1
            else:
                offset = None

            if neg_index < 0:
                response = np.squeeze(y[:,:neg_index])

        return response, case_weights, offset

    def _saturated_score(self,
                         predictions,
                         response,
                         case_weights=None):
        raise NotImplementedError

class lagrange_mixin(base_mixin):

    """
    A regularized estimator where atom constructed
    is assumed to be in Lagrange form. Allows 
    for ElasticNet type term.

    The atom constructor can use the gradient
    of the loss at 0 in order to construct
    the appropriate atom -- commonly the 
    Lagrange parameter is taken to be some multiple
    in [0,1] of the dual norm of this vector.
    """

    def __init__(self, 
                 atom_constructor, 
                 atom_params, 
                 case_weights=False,
                 offset=False,
                 enet_alpha=0.,
                 coef=1., 
                 score_method='deviance',
                 unpenalized=False,
                 solve_args={},
                 initial=None):

        """

        Parameters
        ----------

        atom_constructor : callable

            Atom constructor that is to to be used as a regularizer.
            Final atom will be constructed as `atom_constructor(**atom_params)`.

        atom_params : dict
            Dictionary of arguments to construct atom.

        offset : bool
            While `y` include an offset column when fitting?

        case_weights : bool
            Will `y` include a case weight column when fitting?

        enet_alpha : float
            Weighting in [0, 1] between ridge and regularizer.

        coef : float
            Scaling to be put in front of loss.

        score_method : str
            Which score to use as default `score`?
            One of ['deviance', 'mean_deviance', 'R2']

        unpenalized : bool
            If any features are unpenalized,
            the null solution with
            `lagrange=np.inf` requires solving a 
            problem. 

        solve_args : dict
            Keyword arguments passed to `simple_problem.solve`

        Notes
        -----

        If `case_weights` or `offset` is True, then the `fit`
        method should pass a 2-d array for `y` rather than just a
        response. If only one of these is not None, then the
        quantity is assumed to be the last column.  If both are
        not None, then the 2nd to last are to be the case weights
        and the last is to be the offset.

        The `offset` argument, if not None
        is *subtracted* from the linear predictor
        before evaluating the loss -- this sign is different
        than behavior in e.g. `R` where it is *added* instead.

        """

        self.atom_constructor = atom_constructor
        self.atom_params = atom_params
        self.case_weights = case_weights
        self.coef = coef
        self.offset = offset
        self.score_method = score_method
        self.enet_alpha = enet_alpha
        self.unpenalized = unpenalized
        self.solve_args = solve_args
        self.initial = initial
        
    def fit(self, X, y):
        """
        Fit a regularized regression estimator.

        Parameters
        ----------

        X : np.ndarray((n, p))
            Feature matrix.

        y : np.ndarray(n)
            Response vector.

        Returns
        -------

        self

        """

        self._loglike = loglike = self._loglike_factory(X, y)

        # with unpenalized parameters possible,
        # this may be best found by solving a problem with an atom with lagrange=np.inf
        # this could get expensive though

        null_grad = loglike.smooth_objective(np.zeros(loglike.shape), 'grad')
        atom_ = self._construct_atom(null_grad)
        if self.unpenalized:
            null_grad = self._fit_null_soln(loglike, atom_)
        atom_ = self._construct_atom(null_grad)
        problem = simple_problem(loglike, atom_)
        if self.initial is not None:
            problem.coefs[:] = self.initial
        self._coefs = problem.solve(**self.solve_args)

        return self

    def _fit_null_soln(self, loglike, atom):
        old_lagrange = atom.lagrange
        atom.lagrange = np.inf
        problem = simple_problem(loglike, atom)
        null_soln = problem.solve()
        null_grad = loglike.smooth_objective(null_soln, 'grad')
        return null_grad

    def _construct_atom(self, score):
        atom_ = self.atom_constructor(score, **self.atom_params)
        if self.enet_alpha != 0:
            # reweight penalty and add elastic net
            ridge_term = identity_quadratic(self.enet_alpha * atom_.lagrange, 0, 0, 0)
            atom_.lagrange *= (1 - self.enet_alpha)
            atom_.quadratic = ridge_term
        return atom_
            
class classifier_mixin(base_mixin):

    def __init__(self, 
                 atom_constructor, 
                 atom_params, 
                 case_weights=False,
                 coef=1., 
                 quadratic=None,
                 score_method='accuracy',
                 initial=None):

        """

        Parameters
        ----------

        atom_constructor : callable

            Atom constructor that is to to be used as a regularizer.
            Final atom will be constructed as `atom_constructor(**atom_params)`.

        atom_params : dict
            Dictionary of arguments to construct atom.

        case_weights : bool
            Will `y` include a case weight column when fitting?

        coef : float
            Scaling to be put in front of loss.

        score_method : str
            Which score to use as default `score`?
            One of ['accuracy']

        Notes
        -----

        If `case_weights` is True, then the `fit`
        method should pass at least 2-d array for `y` rather than just a
        response. The last column will be taken to be
        `case_weights`.

        """

        base_mixin.__init__(self,
                            atom_constructor, 
                            atom_params, 
                            case_weights=case_weights,
                            coef=coef,
                            quadratic=quadratic,
                            score_method=score_method,
                            initial=initial)

    def predict(self, X):
        """
        Predict new labels in setting.

        Parameters
        ----------

        X : np.ndarray((n, p))
            Feature matrix.

        Returns
        -------

        labels : np.ndarray(n)
            Predictions from classification model.

        """
        return X.dot(self._coefs) > 0

    def predict_proba(self, X):
        """
        Predict new probabilities in classification setting.

        Parameters
        ----------

        X : np.ndarray((n, p))
            Feature matrix.

        Returns
        -------

        probs : np.ndarray(n)
            Predictions from classification model.

        """
        raise NotImplementedError

class classifier_lagrange_mixin(classifier_mixin, lagrange_mixin):

    def __init__(self, 
                 atom_constructor, 
                 atom_params, 
                 case_weights=False,
                 enet_alpha=0.,
                 coef=1., 
                 score_method='accuracy',
                 initial=None):

        """

        Parameters
        ----------

        atom_constructor : callable

            Atom constructor that is to to be used as a regularizer.
            Final atom will be constructed as `atom_constructor(**atom_params)`.

        atom_params : dict
            Dictionary of arguments to construct atom.

        case_weights : bool
            Will `y` include a case weight column when fitting?

        enet_alpha : float
            Weighting in [0, 1] between ridge and regularizer.

        coef : float
            Scaling to be put in front of loss.

        score_method : str
            Which score to use as default `score`?
            One of ['accuracy']

        Notes
        -----

        If `case_weights` is True, then the `fit`
        method should pass at least 2-d array for `y` rather than just a
        response. The last column will be taken to be
        `case_weights`.

        """

        lagrange_mixin(self, 
                       atom_constructor, 
                       atom_params, 
                       case_weights=case_weights,
                       offset=offset,
                       enet_alpha=enet_alpha,
                       coef=coef, 
                       score_method=score_method,
                       initial=initial)
                       

class survival_mixin(base_mixin):
    pass

class survival_mixin_lagrange(survival_mixin, lagrange_mixin):

    def __init__(self, 
                 atom_constructor, 
                 atom_params, 
                 case_weights=False,
                 offset=False,
                 enet_alpha=0.,
                 coef=1., 
                 score_method='deviance',
                 initial=None):

        """

        Parameters
        ----------

        atom_constructor : callable

            Atom constructor that is to to be used as a regularizer.
            Final atom will be constructed as `atom_constructor(**atom_params)`.

        atom_params : dict
            Dictionary of arguments to construct atom.

        offset : bool
            While `y` include an offset column when fitting?

        case_weights : bool
            Will `y` include a case weight column when fitting?

        enet_alpha : float
            Weighting in [0, 1] between ridge and regularizer.

        coef : float
            Scaling to be put in front of loss.

        score_method : str
            Which score to use as default `score`?
            One of ['deviance', 'mean_deviance', 'R2', 'C-index']

        Notes
        -----

        If `case_weights` or `offset` is True, then the `fit`
        method should pass a 2-d array for `y` rather than just a
        response. If only one of these is not None, then the
        quantity is assumed to be the last column.  If both are
        not None, then the 2nd to last are to be the case weights
        and the last is to be the offset.

        The `offset` argument, if not None
        is *subtracted* from the linear predictor
        before evaluating the loss -- this sign is different
        than behavior in e.g. `R` where it is *added* instead.

        """

        lagrange_mixin.__init__(self, 
                                atom_constructor, 
                                atom_params, 
                                case_weights=case_weights,
                                offset=offset,
                                enet_alpha=enet_alpha,
                                coef=coef, 
                                score_method=score_method,
                                initial=initial)
                       

### Final mixins

if have_sklearn:

    class sklearn_regression_lagrange(lagrange_mixin, BaseEstimator, RegressorMixin): 
        pass

    class sklearn_regression(base_mixin, BaseEstimator, RegressorMixin): 
        pass

    class sklearn_classifier(classifier_mixin, BaseEstimator, ClassifierMixin):
        pass

    class sklearn_classifier_lagrange(classifier_lagrange_mixin, BaseEstimator, ClassifierMixin):
        pass

    class sklearn_survival(survival_mixin, BaseEstimator, RegressorMixin):
        pass

    class sklearn_survival_lagrange(classifier_mixin, BaseEstimator, RegressorMixin):
        pass




"""
This module contains the implementation of block norms, i.e.
l1/l*, linf/l* norms. These are used in multiresponse LASSOs.

"""
from __future__ import print_function, division, absolute_import

import warnings
from copy import copy

import numpy as np

from . import seminorms
from ..identity_quadratic import identity_quadratic
from ..problems.composite import smooth_conjugate
from ..objdoctemplates import objective_doc_templater
from ..doctemplates import (doc_template_user, doc_template_provider)
from ..atoms import _work_out_conjugate
from .block_norms import l1_l2
from .sparse_group_lasso import _gauge_function_dual_strong, _inside_set_strong

# for the docstring, we need l1norm
l1norm = seminorms.l1norm

@objective_doc_templater()
class sparse_group_block(l1_l2):

    objective_template = r"""w_1\|%(var)s\|_{1,1} + w_1\|%(var)s\|_{1,2}"""
    objective_vars = l1_l2.objective_vars.copy()
    objective_vars['var'] = 'B'
    objective_vars['normklass'] = 'sparse_group_block'
    objective_vars['dualnormklass'] = 'sparse_group_block_dual'
    objective_vars['initargs'] = '(5, 4), 1, 2'
    objective_vars['shape'] = r'n \times p'

    def __init__(self, 
                 shape,
                 l1_weight,
                 l2_weight,
                 lagrange=None,
                 bound=None,
                 offset=None,
                 quadratic=None,
                 initial=None):

        l1_l2.__init__(self,
                       shape,
                       lagrange=lagrange,
                       bound=bound,
                       offset=offset,
                       quadratic=quadratic,
                       initial=initial)

        self.l1_weight = l1_weight
        self.l2_weight = l2_weight

    @doc_template_user
    def lagrange_prox(self, arg, lipschitz=1, lagrange=None):
        arg = arg.reshape(self.shape)
        lagrange = seminorms.seminorm.lagrange_prox(self, arg, lipschitz, lagrange)
        return _lagrange_prox(arg, 
                              lagrange * self.l1_weight / lipschitz,
                              lagrange * self.l2_weight / lipschitz)

    @doc_template_user
    def bound_prox(self, arg, bound=None):
        raise NotImplementedError('sparse_group_block bound form not implemented')

    @doc_template_user
    def constraint(self, x):
        x = x.reshape(self.shape)
        l1_norms = np.fabs(x).sum()
        l2_norms = np.sqrt(np.sum(x**2), 1).sum()
        norm_sum = self.l1_weight * l1_norms + self.l2_weight * l2_norms
        if norm_sum <= self.bound * (1 + self.tol):
            return 0
        return np.inf

    @doc_template_user
    def seminorm(self, x, lagrange=None, check_feasibility=False):
        x = x.reshape(self.shape)
        lagrange = seminorms.seminorm.seminorm(self, x, lagrange=lagrange,
                                               check_feasibility=check_feasibility)
        l1_norms = np.fabs(x).sum()
        l2_norms = np.sqrt(np.sum(x**2, 1)).sum()
        return lagrange * (self.l1_weight * l1_norms +
                           self.l2_weight * l2_norms)

    @doc_template_user
    def get_conjugate(self):
        if self.quadratic.coef == 0:
            offset, outq = _work_out_conjugate(self.offset, 
                                               self.quadratic)

            cls = sparse_group_block_pairs[self.__class__]
            conj_atom = self.atom.conjugate

            atom = cls(self.shape, 
                       self.l1_weight,
                       self.l2_weight,
                       offset=offset,
                       lagrange=conj_atom.lagrange,
                       bound=conj_atom.bound,
                       quadratic=outq)

        else:
            atom = smooth_conjugate(self)
        self._conjugate = atom
        self._conjugate._conjugate = self
        return self._conjugate
    conjugate = property(get_conjugate)

    def __copy__(self):
         return self.__class__(self.shape,
                               self.l1_weight,
                               self.l2_weight,
                               quadratic=self.quadratic,
                               initial=self.coefs,
                               bound=copy(self.bound),
                               lagrange=copy(self.lagrange),
                               offset=copy(self.offset))

    def terms(self, arg):
        """
        Return the args that are summed
        in computing the seminorm.

        >>> import regreg.api as rr
        >>> groups = [1,1,2,2,2]
        >>> penalty = rr.group_lasso(groups, lagrange=1.)
        >>> arg = [2,4,5,3,4]
        >>> list(penalty.terms(arg)) # doctest: +ELLIPSIS
        [6.3245..., 12.2474...]
        >>> penalty.seminorm(arg) # doctest: +ELLIPSIS
        18.5720...
        >>> np.sqrt((2**2 + 4**2)*2), np.sqrt((5**2 + 3**2 + 4**2) * 3.) # doctest: +ELLIPSIS
        (6.3245..., 12.2474...)
        >>> np.sqrt((2**2 + 4**2)*2) + np.sqrt((5**2 + 3**2 + 4**2) * 3.) # doctest: +ELLIPSIS
        18.5720...
        
        """
        terms = (np.fabs(arg).sum(1) * self.l1_weight + 
                 np.sqrt((arg**2).sum(1)) * self.l1_weight)
        return terms

class sparse_group_block_dual(sparse_group_block):

    objective_template = r"""\|%(var)s\|_{w_1,w_2,\text{block}}"""
    objective_vars = l1_l2.objective_vars.copy()
    objective_vars['var'] = 'B'
    objective_vars['normklass'] = 'sparse_group_block_dual'
    objective_vars['dualnormklass'] = 'sparse_group_block'
    objective_vars['initargs'] = '(5, 4), 1, 2'
    objective_vars['shape'] = r'n \times p'

    def __init__(self, 
                 shape,
                 l1_weight,
                 l2_weight,
                 lagrange=None,
                 bound=None,
                 offset=None,
                 quadratic=None,
                 initial=None):

        l1_l2.__init__(self,
                       shape,
                       lagrange=lagrange,
                       bound=bound,
                       offset=offset,
                       quadratic=quadratic,
                       initial=initial)

        self.l1_weight = l1_weight
        self.l2_weight = l2_weight

    @doc_template_user
    def lagrange_prox(self, arg, lipschitz=1, lagrange=None):
        raise NotImplementedError('sparse_group_block Lagrange form not implemented')

    @doc_template_user
    def bound_prox(self, arg, bound=None):
        arg = arg.reshape(self.shape)
        bound = seminorms.seminorm.bound_prox(self, arg, bound)
        _prox = _lagrange_prox(arg, 
                               bound * self.l1_weight,
                               bound * self.l2_weight)
        return arg - _prox

    @doc_template_user
    def constraint(self, x):
        x = x.reshape(self.shape)
        dual_norm = _gauge_function_dual(x,
                                         self.l1_weight,
                                         self.l2_weight)
        if dual_norm <= self.bound * (1 + self.tol):
            return 0
        return np.inf

    @doc_template_user
    def seminorm(self, x, lagrange=None, check_feasibility=False):
        x = x.reshape(self.shape)
        lagrange = seminorms.seminorm.seminorm(self, x, lagrange=lagrange,
                                               check_feasibility=check_feasibility)
        return lagrange * _gauge_function_dual(x,
                                               self.l1_weight,
                                               self.l2_weight)

    @doc_template_user
    def get_conjugate(self):
        if self.quadratic.coef == 0:
            offset, outq = _work_out_conjugate(self.offset, 
                                               self.quadratic)

            cls = sparse_group_block_pairs[self.__class__]
            conj_atom = self.atom.conjugate

            atom = cls(self.shape, 
                       self.l1_weight,
                       self.l2_weight,
                       offset=offset,
                       lagrange=conj_atom.lagrange,
                       bound=conj_atom.bound,
                       quadratic=outq)

        else:
            atom = smooth_conjugate(self)
        self._conjugate = atom
        self._conjugate._conjugate = self
        return self._conjugate
    conjugate = property(get_conjugate)

    def terms(self, arg):
         """
         Return the args that are maximized
         in computing the seminorm.
         
         >>> import regreg.api as rr
         >>> groups = [1,1,2,2,2]
         >>> penalty = rr.group_lasso_dual(groups, lagrange=1.)
         >>> arg = [2,4,5,3,4]
         >>> list(penalty.terms(arg)) # doctest: +ELLIPSIS
         [3.1622..., 4.0824...]
         >>> np.sqrt((2**2 + 4**2)/2), np.sqrt((5**2 + 3**2 + 4**2) / 3.) # doctest: +ELLIPSIS
         (3.1622..., 4.0824...)
         >>> penalty.seminorm(arg) # doctest: +ELLIPSIS
         4.0824...

         """
         return np.array([_gauge_function_dual_strong(arg[i],
                                                      self.l1_weight,
                                                      self.l2_weight)[0] for i in range(arg.shape[0])])

# fast Lagrange prox

def _lagrange_prox(arg, l1_weight, l2_weight):
    soft_thresh = np.sign(arg) * np.maximum(np.fabs(arg) - l1_weight, 0)
    norms = np.sqrt(np.sum(soft_thresh**2, 1))
    norm_factors = np.maximum(norms - l2_weight, 0) / (norms + (norms == 0))
    return soft_thresh * norm_factors[:, None]
    
# for computing dual norm

def _inside_set(point, bound, l1_weight, l2_weight):

    prox_point = _lagrange_prox(point,
                                bound * l1_weight,
                                bound * l2_weight)
    if np.linalg.norm(prox_point) > max(np.linalg.norm(point), 1) * 1.e-10:
        return False
    return True

def _gauge_function_dual(point,
                         l1_weight,
                         l2_weight,
                         tol=1.e-6,
                         max_iter=50): 

     """
     Work out dual norm of sparse group LASSO by binary search.

     NOTE: will have problems if the atom has infinite feature weights
     """

     point = np.asarray(point)

     # find upper and lower bounds

     lower, upper = 1., 1.
     point_inside = _inside_set(point,
                                lower, 
                                l1_weight,
                                l2_weight)
     
     iter = 0
     if point_inside:
          # gauge is upper bounded by 1
          # find a lower bound

          while True:
               lower = lower / 2
               if not _inside_set(point,
                                  lower,
                                  l1_weight,
                                  l2_weight):
                    break
               else:
                    upper = lower

               iter += 1
               if iter == max_iter:
                    return 0
     else:
          # gauge is lower bounded by 1
          # find an upper bound

          while True:
               upper *= 2
               if _inside_set(point,
                              upper,
                              l1_weight,
                              l2_weight):
                    break
               else:
                    lower = upper

               iter += 1
               if iter == max_iter:
                    return np.inf

     # binary search

     assert (not _inside_set(point,
                             lower,
                             l1_weight,
                             l2_weight))

     assert _inside_set(point,
                        upper,
                        l1_weight,
                        l2_weight)

     while (upper - lower) > tol * 0.5 * (upper + lower):
         mid = 0.5 * (upper + lower)
         if _inside_set(point,
                        mid,
                        l1_weight,
                        l2_weight):
             upper = mid
         else:
             lower = mid
     return mid

sparse_group_block_pairs = {}
for n1, n2 in [(sparse_group_block, sparse_group_block_dual)]:
    sparse_group_block_pairs[n1] = n2
    sparse_group_block_pairs[n2] = n1




import gc
from warnings import warn
import numpy as np

from ..affine import astransform, power_L
from ..smooth import affine_smooth, sum as smooth_sum
from ..problems.simple import simple_problem

def subsample_columns(X, columns):

    X = astransform(X)
    cols = np.zeros((len(columns),) + X.output_shape)
    indicator = np.zeros(X.input_shape[0]) # assuming 1-dimensional input and output shape here
    for i, col in enumerate(columns):
        indicator[col] = 1 # 1-hot vector
        cols[i] = X.dot(indicator)
        indicator[col] = 0 # back to 0 vector
    return cols.T

class grouped_path(object):

    BIG = 1e12 # lagrange parameter for finding null solution

    def __init__(self, 
                 saturated_loss,
                 X, 
                 lasso_weights,
                 elastic_net_param=None,
                 alpha=1.  # elastic net mixing -- 1 is LASSO
                 ):
        raise NotImplementedError

    # used for warm start path

    def full_loss(self):
        loss = affine_smooth(self.saturated_loss,
                             self.X)
        loss.shape = self.X.input_shape
        return loss

    def enet_loss(self, 
                  lagrange,
                  candidate_set):
        raise NotImplementedError

    # methods potentially overwritten in subclasses for I/O considerations

    def subsample_columns(self, 
                          X, 
                          columns):
        """
        Extract columns of X into ndarray or
        regreg transform
        """
        return subsample_columns(X, 
                                 columns)

    def full_gradient(self, 
                      saturated_loss, 
                      linear_predictor):
        """
        Gradient of saturated loss composed with self.X
        """
        saturated_grad = saturated_loss.smooth_objective(linear_predictor, 'grad')
        return self.X.T.dot(saturated_grad)

    def linpred(self,
                coef,
                X,
                test):
        """
        Form linear predictor for given set of coefficients.

        Typical use case will have coef of shape `(l,p)` for
        univariate response regressions and `(l,q,p)` for
        multiple response regressions (e.g. multinomial with
        `q` classes) where `l` is the number of Lagrange
        parameters.

        Parameters
        ----------

        coef : ndarray
            A set of coefficients. 

        X : object
            Representation of design matrix, usually `self.X` -- subclasses
            may not assume this is actually an array.

        test : index
            Indices of X.dot(coef.T) to return.

        Returns
        -------

        linpred : Matrix of linear predictors.

        """
        coef = np.asarray(coef)
        if coef.ndim > 1:
            coefT = coef.T
            old_shape = coefT.shape
            coefT = coefT.reshape((coefT.shape[0], -1))
            linpred = X.dot(coefT)[test]
            linpred = linpred.reshape((linpred.shape[0],) + old_shape[1:])
            return linpred
        return X.dot(coef.T)[test]

    # method potentially overwritten in subclasses for penalty considerations

    def subsample(self,
                  case_idx):
        '''

        Create a new path, by subsampling
        cases of `self.saturated_loss`.

        Case weights are computed
        with `self.saturated_loss.subsample`.

        Parameters
        ----------

        case_idx : index
            An index-like object used 
            to specify which cases to include
            in the subsample.

        Returns
        -------

        subsample_path : path object
            A path object with a modified smooth part
            reflecting the subsampling.

        '''
        raise NotImplementedError

    def check_KKT(self,
                  grad_solution,
                  solution,
                  lagrange,
                  penalty=None):

        '''

        Check KKT conditions over
        the groups in the path.
        Returns boolean indicating
        which groups are failing the KKT conditions
        (these could be `active` groups or
        `inactive` groups).

        Parameters
        ----------

        grad_solution : ndarray
             Candidate for gradient of smooth loss at 
             Lagrange value `lagrange`.

        solution : ndarray
             Candidate for solution to problem 
             Lagrange value `lagrange`.

        lagrange : float
             Lagrange value for penalty

        penalty : object (optional)
             A weighted L1 penalty. If None, defaults
             to `self.penalty`.

        '''
        raise NotImplementedError

    def strong_set(self,
                   lagrange_cur,
                   lagrange_new,
                   grad_solution):
        raise NotImplementedError

    def active_set(self, 
                   solution):
        raise NotImplementedError

    def restricted_penalty(self, 
                           subset):
        raise NotImplementedError

    def solve_subproblem(self, 
                         candidate_set, 
                         lagrange_new, 
                         **solve_args):
        raise NotImplementedError

    def enet_grad(self,
                  solution,
                  penalized, # boolean
                  lagrange_new,
                  subset=None):
        raise NotImplementedError

    def updated_ever_active(self,
                            index_obj):
        raise NotImplementedError

def default_lagrange_sequence(penalty,
                              null_solution,
                              lagrange_proportion=0.05,
                              nstep=100,
                              alpha=1):
    dual = penalty.conjugate
    lagrange_max = dual.seminorm(null_solution, lagrange=1)
    return (lagrange_max * np.exp(np.linspace(np.log(lagrange_proportion), 
                                              0, 
                                              nstep)))[::-1] / alpha
   
def strong_rules(path_obj,
                 lagrange_seq,
                 initial_data, # (solution, grad) pair
                 inner_tol=1.e-5,
                 verbose=False,
                 initial_step=None,
                 check_active=False,
                 max_tries=20):

    # take a guess at the inverse step size
    if initial_step is None:
        _lipschitz = power_L(path_obj.X, max_its=50)
        final_step = 1000. / _lipschitz 
    else:
        final_step = initial_step

    # gradient of restricted elastic net at lambda_max

    solution, grad_solution = initial_data

    solution = solution.copy()
    grad_solution = grad_solution.copy()

    linear_predictor = path_obj.X.dot(solution)
    ever_active = path_obj.updated_ever_active([])

    dev_solution = path_obj.saturated_loss.smooth_objective(linear_predictor, 'func')
    solutions = [solution.T.copy()]
    dev_explained = [dev_solution]

    all_failing = np.zeros(path_obj.group_shape, np.bool)
    subproblem_set = path_obj.updated_ever_active([])

    for lagrange_new, lagrange_cur in zip(lagrange_seq[1:], 
                                          lagrange_seq[:-1]):
        tol = inner_tol
        num_tries = 0
        debug = False
        coef_stop = True

        while True:

            subproblem_set = sorted(set(subproblem_set + path_obj.updated_ever_active(all_failing)))

            (final_step, 
             subproblem_grad, 
             subproblem_soln,
             subproblem_linpred,
             subproblem_vars) = path_obj.solve_subproblem(solution, # for warm start
                                                          subproblem_set,
                                                          lagrange_new,
                                                          tol=tol,
                                                          start_step=final_step,
                                                          debug=debug and verbose,
                                                          coef_stop=coef_stop)

            saturated_grad = path_obj.saturated_loss.smooth_objective(subproblem_linpred, 'grad')
            # as subproblem always contains ever active, 
            # rest of solution should be 0
            solution[subproblem_vars] = subproblem_soln

            # strong rules step
            # a set of group ids

            strong, strong_idx, strong_vars = path_obj.strong_set(lagrange_cur * path_obj.alpha, 
                                                                  lagrange_new * path_obj.alpha, 
                                                                  grad_solution)
            strong_enet_grad = path_obj.enet_grad(solution,
                                                  path_obj._penalized_vars,
                                                  lagrange_new,
                                                  subset=strong_vars)
            strong_soln = solution[strong_vars]
            X_strong = path_obj.subsample_columns(path_obj.X, 
                                                  strong_vars)
            strong_grad = (X_strong.T.dot(saturated_grad) +
                           strong_enet_grad)
            strong_penalty = path_obj.restricted_penalty(strong_vars)
            (strong_A,
             strong_I_ranks) = path_obj.check_KKT(strong_grad, 
                                                  strong_soln, 
                                                  path_obj.alpha * lagrange_new, 
                                                  penalty=strong_penalty)

            if check_active:
                strong_failing = strong_A + (strong_I_ranks >= 0)
            else:
                strong_failing = strong_I_ranks >= 0

            if np.any(strong_failing):
                delta = np.zeros(path_obj.group_shape, np.bool)
                delta[strong_idx] = strong_failing
                all_failing += delta 
            else:
                enet_grad = path_obj.enet_grad(solution, 
                                               path_obj._penalized_vars,
                                               lagrange_new)
                grad_solution[:] = (path_obj.full_gradient(path_obj.saturated_loss, 
                                                           subproblem_linpred) + 
                                    enet_grad)
                failing_A, failing_I_ranks = path_obj.check_KKT(grad_solution, 
                                                                solution, 
                                                                path_obj.alpha * lagrange_new)

                if check_active:
                    all_failing = failing_A + (failing_I_ranks >= 0)
                else:
                    all_failing = failing_I_ranks >= 0

                if not all_failing.sum():
                    path_obj.ever_active = path_obj.updated_ever_active(path_obj.active_set(solution))
                    linear_predictor[:] = subproblem_linpred
                    break
                else:
                    if verbose:
                        print('failing:', np.nonzero(all_failing)[0])

            num_tries += 1

            tol /= 2.
            if num_tries % 5 == 0:

                solution[subproblem_vars] = subproblem_soln
                solution[~subproblem_vars] = 0

                enet_grad = path_obj.enet_grad(solution, 
                                               path_obj._penalized_vars,
                                               lagrange_new)
                grad_solution[:] = (path_obj.full_gradient(path_obj.saturated_loss, 
                                                           subproblem_linpred) + 
                                    enet_grad)
                debug = True
                tol = inner_tol
                if num_tries >= max_tries:
                    warn('convergence of active set not achieved for lagrange=%0.4e, moving on' % lagrange_new)
                    break

            subproblem_set = sorted(set(subproblem_set + path_obj.updated_ever_active(all_failing)))

        solutions.append(solution.T.copy())
        dev_explained.append(path_obj.saturated_loss.smooth_objective(linear_predictor, mode='func'))
        gc.collect()

        if verbose:
            print({'lagrange':lagrange_new,
                   'sparsity':(solution != 0).sum(),
                   'dev explained':1. - dev_explained[-1] / dev_explained[0],
                   'point on path':list(lagrange_seq).index(lagrange_new),
                   'penalty':path_obj.penalty.seminorm(solution, lagrange=1),
                   'dual penalty':path_obj.penalty.conjugate.seminorm(grad_solution, lagrange=1)})

    dev_explained = np.array(dev_explained)
    output = {'dev explained': 1 - dev_explained[-1] / dev_explained[0],
              'lagrange': lagrange_seq,
              'beta':np.array(solutions)}

    return output

def warm_start(path_obj,
               lagrange_seq,
               initial_data,
               inner_tol=1.e-5,
               verbose=False,
               initial_step=None,
               solve_args={}):
    
    solution, _ = initial_data

    # basic setup

    loss = path_obj.full_loss()
    all_vars = np.ones(path_obj.penalty.shape, np.bool)
    penalty = path_obj.restricted_penalty(None)

    # take a guess at the inverse step size
    if initial_step is None:
        _lipschitz = power_L(path_obj.X, max_its=50)
        path_obj.final_step = 1000. / _lipschitz 
    else:
        path_obj.final_step = initial_step

    objective = []
    solutions = []
    
    for lagrange in lagrange_seq:
        enet_loss = path_obj.enet_loss(lagrange)
        total_loss = smooth_sum([loss, enet_loss])
        penalty.lagrange = lagrange * path_obj.alpha

        problem = simple_problem(total_loss, penalty)
        problem.coefs[:] = solution
        solve_args['tol'] = inner_tol
        solve_args['start_step'] = path_obj.final_step
        solution = problem.solve(**solve_args)

        objective.append(loss.smooth_objective(solution, mode='func'))
        solutions.append(solution.T.copy())
        grad_solution = total_loss.smooth_objective(solution, 'grad')
        path_obj.final_step = problem.final_step

        if verbose:
            print({'lagrange':lagrange,
                   'sparsity':(solution != 0).sum(),
                   'dev explained':1. - objective[-1] / objective[0],
                   'point on path':list(lagrange_seq).index(lagrange),
                   'penalty':path_obj.penalty.seminorm(solution, lagrange=1),
                   'dual penalty':path_obj.penalty.conjugate.seminorm(grad_solution, lagrange=1)})

    objective = np.array(objective)
    output = {'devratio': 1 - objective / objective.max(),
              'lagrange': lagrange_seq,
              'beta':np.array(solutions)}

    return output





"""
This module contains the implementation operator and nuclear norms, used in 
matrix completion problems and other low-rank factorization
problems.

"""
from copy import copy

import numpy as np

from ..atoms import atom, _work_out_conjugate
from .seminorms import conjugate_seminorm_pairs, seminorm, l1norm, supnorm
from .cones import cone, conjugate_cone_pairs
from .projl1_cython import projl1, projl1_epigraph
from .piecewise_linear import find_solution_piecewise_linear_c

from ..objdoctemplates import objective_doc_templater
from ..doctemplates import (doc_template_user, doc_template_provider)

# for doctests
from ..identity_quadratic import identity_quadratic

@objective_doc_templater()
class svd_atom(seminorm):

    objective_vars = seminorm.objective_vars.copy()
    objective_vars['var'] = 'B'
    objective_vars['normklass'] = 'nuclear_norm'
    objective_vars['dualnormklass'] = 'operator_norm'
    objective_vars['initargs'] = '(5,4)'
    objective_vars['shape'] = r'{n \times p}'


@objective_doc_templater()
class nuclear_norm(svd_atom):

    """
    The nuclear norm
    """
    prox_tol = 1.0e-10

    objective_template = r"""\|%(var)s\|_*"""
    objective_vars = svd_atom.objective_vars.copy()

    @doc_template_user
    def seminorm(self, X, check_feasibility=False,
                 lagrange=None):
        lagrange = seminorm.seminorm(self, X, lagrange=lagrange,
                                 check_feasibility=check_feasibility)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        return lagrange * np.fabs(D).sum()

    @doc_template_user
    def constraint(self, X, bound=None):
        bound = seminorm.constraint(self, X, bound=bound)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        inbox = np.fabs(D).sum() <= bound * (1 + self.tol)
        if inbox:
            return 0
        else:
            return np.inf

    @doc_template_user
    def lagrange_prox(self, X,  lipschitz=1, lagrange=None):
        lagrange = svd_atom.lagrange_prox(self, X, lipschitz, lagrange)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        l1atom = l1norm(np.product(self.shape), lagrange=lagrange)
        D_soft_thresholded = l1atom.lagrange_prox(D, lipschitz=lipschitz, lagrange=lagrange)
        keepD = D_soft_thresholded > 0
        U_new, D_new, V_new = U[:,keepD], D_soft_thresholded[keepD], V[keepD]
        return np.dot(U, D_soft_thresholded[:,np.newaxis] * V)

    @doc_template_user
    def bound_prox(self, X, bound=None):
        bound = svd_atom.bound_prox(self, X, bound)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        l1atom = l1norm(np.product(self.shape), bound=bound)
        D_projected = l1atom.bound_prox(D)
        keepD = D_projected > 0
        U_new, D_new, V_new = U[:,keepD], D_projected[keepD], V[keepD]
        return np.dot(U_new, D_new[:,np.newaxis] * V_new)

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return seminorm.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_bound(self):
        return seminorm.get_bound(self)

    @doc_template_user
    def set_bound(self, bound):
        return seminorm.set_bound(self, bound)

    @doc_template_user
    def get_lagrange(self):
        return seminorm.get_lagrange(self)

    @doc_template_user
    def set_lagrange(self, lagrange):
        return seminorm.set_lagrange(self, lagrange)

    @doc_template_user
    def get_conjugate(self):
        return seminorm.get_conjugate(self)

    @doc_template_user
    def get_dual(self):
        return seminorm.dual(self)

@objective_doc_templater()
class operator_norm(svd_atom):

    """
    The operator norm
    """
    prox_tol = 1.0e-10

    objective_template = r"""\|%(var)s\|_{\text{op}}"""
    objective_vars = svd_atom.objective_vars.copy()
    objective_vars['normklass'] = 'operator_norm'
    objective_vars['dualnormklass'] = 'nuclear_norm'

    @doc_template_user
    def seminorm(self, X, lagrange=None, check_feasibility=False):
        lagrange = seminorm.seminorm(self, X, lagrange=lagrange,
                                 check_feasibility=check_feasibility)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        return lagrange * np.max(D)

    @doc_template_user
    def constraint(self, X, bound=None):
        bound = seminorm.constraint(self, X, bound=bound)
        U, D, V = np.linalg.svd(X, full_matrices=0)
        inbox = np.max(D) <= bound * (1 + self.tol)
        if inbox:
            return 0
        else:
            return np.inf

    @doc_template_user
    def lagrange_prox(self, X,  lipschitz=1, lagrange=None):
        lagrange = svd_atom.lagrange_prox(self, X, lipschitz, lagrange)
        U, D, V = np.linalg.svd(X, full_matrices=False)
        supatom = supnorm(np.product(self.shape), lagrange=lagrange)
        D_new = supatom.lagrange_prox(D, lipschitz=lipschitz, lagrange=lagrange)
        return np.dot(U, D_new[:,np.newaxis] * V)

    @doc_template_user
    def bound_prox(self, X, bound=None):
        bound = svd_atom.bound_prox(self, X, bound)
        U, D, V = np.linalg.svd(X, full_matrices=False)
        supatom = supnorm(np.product(self.shape), bound=bound)
        D_new = supatom.bound_prox(D)
        return np.dot(U, D_new[:,np.newaxis] * V)

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return seminorm.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_bound(self):
        return seminorm.get_bound(self)

    @doc_template_user
    def set_bound(self, bound):
        return seminorm.set_bound(self, bound)

    @doc_template_user
    def get_lagrange(self):
        return seminorm.get_lagrange(self)

    @doc_template_user
    def set_lagrange(self, lagrange):
        return seminorm.set_lagrange(self, lagrange)

    @doc_template_user
    def get_conjugate(self):
        return seminorm.get_conjugate(self)

    @doc_template_user
    def get_dual(self):
        return seminorm.dual(self)

@objective_doc_templater()
class svd_cone(cone):

    objective_vars = svd_atom.objective_vars.copy()
    objective_vars['coneklass'] = 'nuclear_norm_epigraph'
    objective_vars['dualconeklass'] = 'nuclear_norm_epigraph_polar'
    objective_vars['shape'] = r'{n \times p + 1}'

    def __copy__(self):
        return self.__class__(copy(self.matrix_shape),
                              offset=copy(self.offset),
                              initial=self.coefs,
                              quadratic=self.quadratic)
    

    def __repr__(self):
        if self.quadratic.iszero:
            return "%s(%s, offset=%s)" % \
                (self.__class__.__name__,
                 repr(self.matrix_shape),
                 str(self.offset))
        else:
            return "%s(%s, offset=%s, quadratic=%s)" % \
                (self.__class__.__name__,
                 repr(self.matrix_shape),
                 str(self.offset),
                 str(self.quadratic))

    def __init__(self, input_shape,
                 offset=None,
                 quadratic=None,
                 initial=None):

        shape = (np.product(input_shape)+1,)
        cone.__init__(self, shape, offset=offset,
                      quadratic=quadratic,
                      initial=initial)
        self.matrix_shape = input_shape

    @doc_template_user
    def get_conjugate(self):
        if self.quadratic.coef == 0:
            offset, outq = _work_out_conjugate(self.offset, 
                                               self.quadratic)
            cls = conjugate_cone_pairs[self.__class__]
            new_atom = cls(self.matrix_shape,
                       offset=offset,
                       quadratic=outq)
        else:
            new_atom = smooth_conjugate(self)
        self._conjugate = new_atom
        self._conjugate._conjugate = self
        return self._conjugate
    conjugate = property(get_conjugate)

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return cone.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_dual(self):
        return cone.dual(self)

@objective_doc_templater()
class nuclear_norm_epigraph(svd_cone):

    objective_template = r"""I^{\infty}(\|%(var)s[:-1]\|_* \leq %(var)s[-1])"""
    objective_vars = svd_cone.objective_vars.copy()
    objective_vars['coneklass'] = 'nuclear_norm_epigraph'
    objective_vars['dualconeklass'] = 'nuclear_norm_epigraph_polar'

    @doc_template_user
    def constraint(self, normX):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        incone = np.fabs(D[1:]).sum() <= (1 + self.tol) * norm
        if incone:
            return 0
        return np.inf

    @doc_template_user
    def cone_prox(self, normX,  lipschitz=1):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)
        newD = np.zeros(D.shape[0]+1)
        newD[-1] = norm
        newD[:-1] = D
        newD = projl1_epigraph(newD)
        result = np.zeros_like(normX)
        result[-1] = newD[-1] 
        newX = np.dot(U, newD[:-1][:,np.newaxis] * V)
        result[:-1] = newX.reshape(-1)
        return result

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return cone.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_dual(self):
        return cone.dual(self)

    @doc_template_user
    def get_conjugate(self):
        return svd_cone.get_conjugate(self)

@objective_doc_templater()
class nuclear_norm_epigraph_polar(svd_cone):
    
    objective_template = r"""I^{\infty}(\|%(var)s[:-1]\|_{op} \leq -%(var)s[-1])"""
    objective_vars = svd_cone.objective_vars.copy()
    objective_vars['coneklass'] = 'nuclear_norm_epigraph_polar'
    objective_vars['dualconeklass'] = 'nuclear_norm_epigraph'

    @doc_template_user
    def constraint(self, normX):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        incone = D.max() <= (1 + self.tol) * (-norm)
        if incone:
            return 0
        return np.inf

    @doc_template_user
    def cone_prox(self, normX,  lipschitz=1):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        newD = np.zeros(D.shape[0]+1)
        newD[-1] = norm
        newD[:-1] = D
        newD = newD - projl1_epigraph(newD)
        result = np.zeros_like(normX)
        result[-1] = newD[-1]
        newX = np.dot(U, newD[:-1][:,np.newaxis] * V)
        result[:-1] = newX.reshape(-1)
        return result

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return cone.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_dual(self):
        return cone.dual(self)

    @doc_template_user
    def get_conjugate(self):
        return svd_cone.get_conjugate(self)

@objective_doc_templater()
class operator_norm_epigraph(svd_cone):
    
    objective_template = r"""I^{\infty}(\|%(var)s[:-1]\|_{op} \leq %(var)s[-1])"""
    objective_vars = svd_cone.objective_vars.copy()
    objective_vars['coneklass'] = 'operator_norm_epigraph'
    objective_vars['dualconeklass'] = 'operator_norm_epigraph_polar'

    @doc_template_user
    def constraint(self, normX):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        incone = D.max() <= (1 + self.tol) * norm
        if incone:
            return 0
        return np.inf

    @doc_template_user
    def cone_prox(self, normX,  lipschitz=1):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        newD = np.zeros(D.shape[0]+1)
        newD[-1] = norm
        newD[:-1] = D
        newD = newD + projl1_epigraph(-newD)
        result = np.zeros_like(normX)
        result[-1] = newD[-1]
        newX = np.dot(U, newD[:-1][:,np.newaxis] * V)
        result[:-1] = newX.reshape(-1)
        return result

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return cone.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_dual(self):
        return cone.dual(self)

    @doc_template_user
    def get_conjugate(self):
        return svd_cone.get_conjugate(self)

@objective_doc_templater()
class operator_norm_epigraph_polar(svd_cone):
    
    objective_template = r"""I^{\infty}(\|%(var)s[:-1]\|_{*} \leq -%(var)s[-1])"""
    objective_vars = svd_cone.objective_vars.copy()
    objective_vars['coneklass'] = 'operator_norm_epigraph_polar'
    objective_vars['dualconeklass'] = 'operator_norm_epigraph'

    @doc_template_user
    def constraint(self, normX):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)

        incone = D.sum() <= (1 + self.tol) * (-norm)
        if incone:
            return 0
        return np.inf

    @doc_template_user
    def cone_prox(self, normX,  lipschitz=1):
        norm = normX[-1]
        X = normX[:-1].reshape(self.matrix_shape)
        U, D, V = np.linalg.svd(X, full_matrices=False)
        newD = np.zeros(D.shape[0]+1)
        newD[-1] = norm
        newD[:-1] = D
        newD = -projl1_epigraph(-newD)
        result = np.zeros_like(normX)
        result[-1] = newD[-1]
        newX = np.dot(U, newD[:-1][:,np.newaxis] * V)
        result[:-1] = newX.reshape(-1)
        return result

    @doc_template_user
    def proximal(self, quadratic, prox_control=None):
        return cone.proximal(self, quadratic, prox_control)

    @doc_template_user
    def get_dual(self):
        return cone.dual(self)

    @doc_template_user
    def get_conjugate(self):
        return svd_cone.get_conjugate(self)

conjugate_svd_pairs = {}
conjugate_svd_pairs[nuclear_norm] = operator_norm
conjugate_svd_pairs[operator_norm] = nuclear_norm

conjugate_seminorm_pairs[nuclear_norm] = operator_norm
conjugate_seminorm_pairs[operator_norm] = nuclear_norm

conjugate_cone_pairs[nuclear_norm_epigraph] = nuclear_norm_epigraph_polar
conjugate_cone_pairs[nuclear_norm_epigraph_polar] = nuclear_norm_epigraph
conjugate_cone_pairs[operator_norm_epigraph] = operator_norm_epigraph_polar
conjugate_cone_pairs[operator_norm_epigraph_polar] = operator_norm_epigraph



from __future__ import print_function, division, absolute_import

from copy import copy

import numpy as np

from ..algorithms import FISTA
from ..smooth import smooth_atom
from .composite import composite
from .simple import simple_problem

from ..identity_quadratic import identity_quadratic

class conjugate(composite):

    def __init__(self, atom, quadratic=None, negate=False, **fit_args):
        
        # we copy the atom because we will modify its quadratic part
        self.atom = copy(atom)

        if quadratic is None:
            quadratic = identity_quadratic(0, 0, 0, 0)
        self.conjugate_quadratic = quadratic

        totalq = self.atom.quadratic + quadratic
        if totalq.coef in [0, None]:
            raise ValueError('quadratic coefficient must be non-zero')

        if isinstance(self.atom, smooth_atom):
            self.problem = simple_problem.smooth(self.atom)
        self.fit_args = fit_args

        #XXX we need a better way to pass around the Lipschitz constant
        # should go in the container class
#         if hasattr(self.atom, "lipschitz"):
#             self.fit_args['perform_backtrack'] = False
#         else:
#             self.fit_args['perform_backtrack'] = True

        self._have_solved_once = False
        self.shape = atom.shape

        # for various purposes,
        # we often want to evalute the conjugate of -x 
        # instead of x

        # negate saves composition with an affine transform

        self.negate = negate

    def smooth_objective(self, x, mode='both', check_feasibility=False):
        """
        Evaluate the conjugate function and/or its gradient

        if mode == 'both', return both function value and gradient
        if mode == 'grad', return only the gradient
        if mode == 'func', return only the function value
        """
        v = -x if self.negate else x

        old_lin = self.conjugate_quadratic.linear_term
        new_lin = -v if old_lin is None else old_lin - v
        self.conjugate_quadratic.linear_term = new_lin
        if isinstance(self.atom, smooth_atom):
            minimizer = self.problem.solve(self.conjugate_quadratic, max_its=5000, **self.fit_args)
        else: # it better have a proximal method!
            minimizer = self.atom.proximal(self.conjugate_quadratic)
        self.conjugate_quadratic.linear_term = old_lin

        # retain a reference
        self.argmin = minimizer

        if mode == 'both':
            val = self.atom.objective(minimizer)
            if not self.negate:
                return -val - self.conjugate_quadratic.objective(minimizer, mode='func') + (v * minimizer).sum(), minimizer
            else:
                return -val - self.conjugate_quadratic.objective(minimizer, mode='func') + (v * minimizer).sum(), -minimizer
        elif mode == 'func':
            val = self.atom.objective(minimizer)
            return -val - self.conjugate_quadratic.objective(minimizer, mode='func') + (v * minimizer).sum()
        elif mode == 'grad':
            return -minimizer if self.negate else minimizer
        raise ValueError("mode incorrectly specified")



from __future__ import print_function, division, absolute_import

from copy import copy

import numpy as np

import scipy.optimize

import regreg.api as rr
from regreg.tests.decorators import set_seed_for_test

@set_seed_for_test()
def test_l1prox():
    '''
    this test verifies that the l1 prox in lagrange form can be solved
    by a primal/dual specification 

    obviously, we don't to solve the l1 prox this way,
    but it verifies that specification is working correctly

    '''

    l1 = rr.l1norm(4, lagrange=0.3)
    ww = np.random.standard_normal(4)*3
    ab = l1.proximal(rr.identity_quadratic(0.5, ww, 0,0))

    l1c = copy(l1)
    l1c.quadratic = rr.identity_quadratic(0.5, ww, None, 0.)
    a = rr.simple_problem.nonsmooth(l1c)
    solver = rr.FISTA(a)
    solver.fit(tol=1.e-10, min_its=100)

    ad = a.coefs

    l1c = copy(l1)
    l1c.quadratic = rr.identity_quadratic(0.5, ww, None, 0.)
    a = rr.dual_problem.fromprimal(l1c)

    # to ensure code is tested
    a.latexify()
    a.quadratic
    a.solve(return_optimum=True)

    solver = rr.FISTA(a)
    solver.fit(tol=1.0e-14, min_its=100)

    ac = a.primal

    np.testing.assert_allclose(ac + 0.1, ab + 0.1, rtol=1.e-4)
    np.testing.assert_allclose(ac + 0.1, ad + 0.1, rtol=1.e-4)

@set_seed_for_test()
def test_l1prox_bound():
    '''
    this test verifies that the l1 prox in bound form can be solved
    by a primal/dual specification 

    obviously, we don't to solve the l1 prox this way,
    but it verifies that specification is working correctly

    '''

    l1 = rr.l1norm(4, bound=2.)
    ww = np.random.standard_normal(4)*2
    ab = l1.proximal(rr.identity_quadratic(0.5, ww, 0, 0))

    l1c = copy(l1)
    l1c.quadratic = rr.identity_quadratic(0.5, ww, None, 0.)
    a = rr.simple_problem.nonsmooth(l1c)
    solver = rr.FISTA(a)
    solver.fit(min_its=100)

    l1c = copy(l1)
    l1c.quadratic = rr.identity_quadratic(0.5, ww, None, 0.)
    a = rr.dual_problem.fromprimal(l1c)
    solver = rr.FISTA(a)
    solver.fit(min_its=100)

    ac = a.primal

    np.testing.assert_allclose(ac + 0.1, ab + 0.1, rtol=1.e-4)


@set_seed_for_test()
def test_lasso():
    '''
    this test verifies that the l1 prox can be solved
    by a primal/dual specification 

    obviously, we don't to solve the l1 prox this way,
    but it verifies that specification is working correctly

    '''

    l1 = rr.l1norm(4, lagrange=2.)
    l1.quadratic = rr.identity_quadratic(0.5, 0, None, 0.)

    X = np.random.standard_normal((10,4))
    Y = np.random.standard_normal(10) + 3
    
    loss = rr.quadratic_loss.affine(X, -Y, coef=0.5)

    p2 = rr.separable_problem.singleton(l1, loss)
    solver2 = rr.FISTA(p2)
    solver2.fit(tol=1.0e-14, min_its=100)


    f = p2.objective
    ans = scipy.optimize.fmin_powell(f, np.zeros(4), ftol=1.0e-12, xtol=1.e-10)

    print(f(solver2.composite.coefs), f(ans))
    np.testing.assert_allclose(ans + 0.1, solver2.composite.coefs + 0.1, rtol=1.e-3)

